"""
Validation utilities for time series forecasting.

This module provides validation functions for time series data and exogenous variables.
"""

from typing import Any, Union, List, Tuple, Optional, Dict
import warnings
import pandas as pd
import numpy as np
from spotforecast2.exceptions import MissingValuesWarning, DataTypeWarning


def check_y(y: Any, series_id: str = "`y`") -> None:
    """
    Validate that y is a pandas Series without missing values.

    This function ensures that the input time series meets the basic requirements
    for forecasting: it must be a pandas Series and must not contain any NaN values.

    Args:
        y: Time series values to validate.
        series_id: Identifier of the series used in error messages. Defaults to "`y`".

    Raises:
        TypeError: If y is not a pandas Series.
        ValueError: If y contains missing (NaN) values.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> from spotforecast2.utils.validation import check_y
        >>>
        >>> # Valid series
        >>> y = pd.Series([1, 2, 3, 4, 5])
        >>> check_y(y)  # No error
        >>>
        >>> # Invalid: not a Series
        >>> try:
        ...     check_y([1, 2, 3])
        ... except TypeError as e:
        ...     print(f"Error: {e}")
        Error: `y` must be a pandas Series with a DatetimeIndex or a RangeIndex. Found <class 'list'>.
        >>>
        >>> # Invalid: contains NaN
        >>> y_with_nan = pd.Series([1, 2, np.nan, 4])
        >>> try:
        ...     check_y(y_with_nan)
        ... except ValueError as e:
        ...     print(f"Error: {e}")
        Error: `y` has missing values.
    """
    if not isinstance(y, pd.Series):
        raise TypeError(
            f"{series_id} must be a pandas Series with a DatetimeIndex or a RangeIndex. "
            f"Found {type(y)}."
        )

    if y.isna().to_numpy().any():
        raise ValueError(f"{series_id} has missing values.")

    return


def check_exog(
    exog: Union[pd.Series, pd.DataFrame],
    allow_nan: bool = True,
    series_id: str = "`exog`",
) -> None:
    """
    Validate that exog is a pandas Series or DataFrame.

    This function ensures that exogenous variables meet basic requirements:
    - Must be a pandas Series or DataFrame
    - If Series, must have a name
    - Optionally warns if NaN values are present

    Args:
        exog: Exogenous variable/s included as predictor/s.
        allow_nan: If True, allows NaN values but issues a warning. If False,
            raises no warning about NaN values. Defaults to True.
        series_id: Identifier of the series used in error messages. Defaults to "`exog`".

    Raises:
        TypeError: If exog is not a pandas Series or DataFrame.
        ValueError: If exog is a Series without a name.

    Warnings:
        MissingValuesWarning: If allow_nan=True and exog contains NaN values.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> from spotforecast2.utils.validation import check_exog
        >>>
        >>> # Valid DataFrame
        >>> exog_df = pd.DataFrame({"temp": [20, 21, 22], "humidity": [50, 55, 60]})
        >>> check_exog(exog_df)  # No error
        >>>
        >>> # Valid Series with name
        >>> exog_series = pd.Series([1, 2, 3], name="temperature")
        >>> check_exog(exog_series)  # No error
        >>>
        >>> # Invalid: Series without name
        >>> exog_no_name = pd.Series([1, 2, 3])
        >>> try:
        ...     check_exog(exog_no_name)
        ... except ValueError as e:
        ...     print(f"Error: {e}")
        Error: When `exog` is a pandas Series, it must have a name.
        >>>
        >>> # Invalid: not a Series/DataFrame
        >>> try:
        ...     check_exog([1, 2, 3])
        ... except TypeError as e:
        ...     print(f"Error: {e}")
        Error: `exog` must be a pandas Series or DataFrame. Got <class 'list'>.
    """
    if not isinstance(exog, (pd.Series, pd.DataFrame)):
        raise TypeError(
            f"{series_id} must be a pandas Series or DataFrame. Got {type(exog)}."
        )

    if isinstance(exog, pd.Series) and exog.name is None:
        raise ValueError(f"When {series_id} is a pandas Series, it must have a name.")

    if not allow_nan:
        if exog.isna().to_numpy().any():
            warnings.warn(
                f"{series_id} has missing values. Most machine learning models "
                f"do not allow missing values. Fitting the forecaster may fail.",
                MissingValuesWarning,
            )

    return


def check_exog_dtypes(
    exog: Union[pd.Series, pd.DataFrame],
    call_check_exog: bool = True,
    series_id: str = "`exog`",
) -> None:
    """
    Check that exogenous variables have valid data types (int, float, category).

    This function validates that the exogenous variables (Series or DataFrame)
    contain only supported data types: integer, float, or category. It issues a
    warning if other types (like object/string) are found, as these may cause
    issues with some machine learning estimators.

    It also strictly enforces that categorical columns must have integer categories.

    Args:
        exog: Exogenous variables to check.
        call_check_exog: If True, calls check_exog() first to ensure basic validity.
            Defaults to True.
        series_id: Identifier used in warning/error messages. Defaults to "`exog`".

    Raises:
        TypeError: If categorical columns contain non-integer categories.

    Warnings:
        DataTypeWarning: If columns with unsupported data types (not int, float, category)
            are found.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> from spotforecast2.utils.validation import check_exog_dtypes
        >>>
        >>> # Valid types (float, int)
        >>> df_valid = pd.DataFrame({
        ...     "a": [1.0, 2.0, 3.0],
        ...     "b": [1, 2, 3]
        ... })
        >>> check_exog_dtypes(df_valid)  # No warning
        >>>
        >>> # Invalid type (object/string)
        >>> df_invalid = pd.DataFrame({
        ...     "a": [1, 2, 3],
        ...     "b": ["x", "y", "z"]
        ... })
        >>> check_exog_dtypes(df_invalid)
        ... # Issues DataTypeWarning about column 'b'
        >>>
        >>> # Valid categorical (with integer categories)
        >>> df_cat = pd.DataFrame({"a": [1, 2, 1]})
        >>> df_cat["a"] = df_cat["a"].astype("category")
        >>> check_exog_dtypes(df_cat)  # No warning
    """
    if call_check_exog:
        check_exog(exog=exog, allow_nan=False, series_id=series_id)

    valid_dtypes = ("int", "Int", "float", "Float", "uint")

    if isinstance(exog, pd.DataFrame):
        unique_dtypes = set(exog.dtypes)
        has_invalid_dtype = False
        for dtype in unique_dtypes:
            if isinstance(dtype, pd.CategoricalDtype):
                try:
                    is_integer = np.issubdtype(dtype.categories.dtype, np.integer)
                except TypeError:
                    # Pandas StringDtype and other non-numpy dtypes will raise TypeError
                    is_integer = False

                if not is_integer:
                    raise TypeError(
                        "Categorical dtypes in exog must contain only integer values. "
                    )
            elif not dtype.name.startswith(valid_dtypes):
                has_invalid_dtype = True

        if has_invalid_dtype:
            warnings.warn(
                f"{series_id} may contain only `int`, `float` or `category` dtypes. "
                f"Most machine learning models do not allow other types of values. "
                f"Fitting the forecaster may fail.",
                DataTypeWarning,
            )

    else:
        dtype_name = str(exog.dtypes)
        if not (dtype_name.startswith(valid_dtypes) or dtype_name == "category"):
            warnings.warn(
                f"{series_id} may contain only `int`, `float` or `category` dtypes. Most "
                f"machine learning models do not allow other types of values. "
                f"Fitting the forecaster may fail.",
                DataTypeWarning,
            )

        if isinstance(exog.dtype, pd.CategoricalDtype):
            if not np.issubdtype(exog.cat.categories.dtype, np.integer):
                raise TypeError(
                    "Categorical dtypes in exog must contain only integer values. "
                )
    return


def get_exog_dtypes(exog: Union[pd.Series, pd.DataFrame]) -> Dict[str, type]:
    """
    Extract and store the data types of exogenous variables.

    This function returns a dictionary mapping column names to their data types.
    For Series, uses the series name as the key. For DataFrames, uses all column names.

    Args:
        exog: Exogenous variable/s (Series or DataFrame).

    Returns:
        Dictionary mapping variable names to their pandas dtypes.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> from spotforecast2.utils.validation import get_exog_dtypes
        >>>
        >>> # DataFrame with mixed types
        >>> exog_df = pd.DataFrame({
        ...     "temp": pd.Series([20.5, 21.3, 22.1], dtype='float64'),
        ...     "day": pd.Series([1, 2, 3], dtype='int64'),
        ...     "is_weekend": pd.Series([False, False, True], dtype='bool')
        ... })
        >>> dtypes = get_exog_dtypes(exog_df)
        >>> dtypes['temp']
        dtype('float64')
        >>> dtypes['day']
        dtype('int64')
        >>>
        >>> # Series
        >>> exog_series = pd.Series([1.0, 2.0, 3.0], name="temperature", dtype='float64')
        >>> dtypes = get_exog_dtypes(exog_series)
        >>> dtypes
        {'temperature': dtype('float64')}
    """
    if isinstance(exog, pd.Series):
        exog_dtypes = {exog.name: exog.dtypes}
    else:
        exog_dtypes = exog.dtypes.to_dict()

    return exog_dtypes


def check_interval(
    interval: Union[List[float], Tuple[float], None] = None,
    ensure_symmetric_intervals: bool = False,
    quantiles: Union[List[float], Tuple[float], None] = None,
    alpha: Optional[float] = None,
    alpha_literal: Optional[str] = "alpha",
) -> None:
    """
    Validate that a confidence interval specification is valid.

    This function checks that interval values are properly formatted and within
    valid ranges for confidence interval prediction.

    Args:
        interval: Confidence interval percentiles (0-100 inclusive).
            Should be [lower_bound, upper_bound]. Example: [2.5, 97.5] for 95% interval.
        ensure_symmetric_intervals: If True, ensure intervals are symmetric
            (lower + upper = 100).
        quantiles: Sequence of quantiles (0-1 inclusive). Currently not validated,
            reserved for future use.
        alpha: Confidence level (1-alpha). Currently not validated, reserved for future use.
        alpha_literal: Name used in error messages for alpha parameter.

    Raises:
        TypeError: If interval is not a list or tuple.
        ValueError: If interval doesn't have exactly 2 values, values out of range (0-100),
            lower >= upper, or intervals not symmetric when required.

    Examples:
        >>> from spotforecast2.utils.validation import check_interval
        >>>
        >>> # Valid 95% confidence interval
        >>> check_interval(interval=[2.5, 97.5])  # No error
        >>>
        >>> # Valid symmetric interval
        >>> check_interval(interval=[2.5, 97.5], ensure_symmetric_intervals=True)  # No error
        >>>
        >>> # Invalid: not symmetric
        >>> try:
        ...     check_interval(interval=[5, 90], ensure_symmetric_intervals=True)
        ... except ValueError as e:
        ...     print("Error: Interval not symmetric")
        Error: Interval not symmetric
        >>>
        >>> # Invalid: wrong number of values
        >>> try:
        ...     check_interval(interval=[2.5, 50, 97.5])
        ... except ValueError as e:
        ...     print("Error: Must have exactly 2 values")
        Error: Must have exactly 2 values
        >>>
        >>> # Invalid: out of range
        >>> try:
        ...     check_interval(interval=[-5, 105])
        ... except ValueError as e:
        ...     print("Error: Values out of range")
        Error: Values out of range
    """
    if interval is not None:
        if not isinstance(interval, (list, tuple)):
            raise TypeError(
                "`interval` must be a `list` or `tuple`. For example, interval of 95% "
                "should be as `interval = [2.5, 97.5]`."
            )

        if len(interval) != 2:
            raise ValueError(
                "`interval` must contain exactly 2 values, respectively the "
                "lower and upper interval bounds. For example, interval of 95% "
                "should be as `interval = [2.5, 97.5]`."
            )

        if (interval[0] < 0.0) or (interval[0] >= 100.0):
            raise ValueError(
                f"Lower interval bound ({interval[0]}) must be >= 0 and < 100."
            )

        if (interval[1] <= 0.0) or (interval[1] > 100.0):
            raise ValueError(
                f"Upper interval bound ({interval[1]}) must be > 0 and <= 100."
            )

        if interval[0] >= interval[1]:
            raise ValueError(
                f"Lower interval bound ({interval[0]}) must be less than the "
                f"upper interval bound ({interval[1]})."
            )

        if ensure_symmetric_intervals and interval[0] + interval[1] != 100:
            raise ValueError(
                f"Interval must be symmetric, the sum of the lower, ({interval[0]}), "
                f"and upper, ({interval[1]}), interval bounds must be equal to "
                f"100. Got {interval[0] + interval[1]}."
            )

    return


def check_predict_input(
    forecaster_name: str,
    steps: Union[int, List[int]],
    is_fitted: bool,
    exog_in_: bool,
    index_type_: type,
    index_freq_: str,
    window_size: int,
    last_window: Optional[Union[pd.Series, pd.DataFrame]],
    last_window_exog: Optional[Union[pd.Series, pd.DataFrame]] = None,
    exog: Optional[
        Union[pd.Series, pd.DataFrame, Dict[str, Union[pd.Series, pd.DataFrame]]]
    ] = None,
    exog_names_in_: Optional[List[str]] = None,
    interval: Optional[List[float]] = None,
    alpha: Optional[float] = None,
    max_step: Optional[int] = None,
    levels: Optional[Union[str, List[str]]] = None,
    levels_forecaster: Optional[Union[str, List[str]]] = None,
    series_names_in_: Optional[List[str]] = None,
    encoding: Optional[str] = None,
) -> None:
    """
    Check all inputs of predict method. This is a helper function to validate
    that inputs used in predict method match attributes of a forecaster already
    trained.

    Args:
        forecaster_name: str
            Forecaster name.
        steps: int, list
            Number of future steps predicted.
        is_fitted: bool
            Tag to identify if the estimator has been fitted (trained).
        exog_in_: bool
            If the forecaster has been trained using exogenous variable/s.
        index_type_: type
            Type of index of the input used in training.
        index_freq_: str
            Frequency of Index of the input used in training.
        window_size: int
            Size of the window needed to create the predictors. It is equal to
            `max_lag`.
        last_window: pandas Series, pandas DataFrame, None
            Values of the series used to create the predictors (lags) need in the
            first iteration of prediction (t + 1).
        last_window_exog: pandas Series, pandas DataFrame, default None
            Values of the exogenous variables aligned with `last_window` in
            ForecasterStats predictions.
        exog: pandas Series, pandas DataFrame, dict, default None
            Exogenous variable/s included as predictor/s.
        exog_names_in_: list, default None
            Names of the exogenous variables used during training.
        interval: list, tuple, default None
            Confidence of the prediction interval estimated. Sequence of percentiles
            to compute, which must be between 0 and 100 inclusive. For example,
            interval of 95% should be as `interval = [2.5, 97.5]`.
        alpha: float, default None
            The confidence intervals used in ForecasterStats are (1 - alpha) %.
        max_step: int, default None
            Maximum number of steps allowed (`ForecasterDirect` and
            `ForecasterDirectMultiVariate`).
        levels: str, list, default None
            Time series to be predicted (`ForecasterRecursiveMultiSeries`
            and `ForecasterRnn).
        levels_forecaster: str, list, default None
            Time series used as output data of a multiseries problem in a RNN problem
            (`ForecasterRnn`).
        series_names_in_: list, default None
            Names of the columns used during fit (`ForecasterRecursiveMultiSeries`,
            `ForecasterDirectMultiVariate` and `ForecasterRnn`).
        encoding: str, default None
            Encoding used to identify the different series (`ForecasterRecursiveMultiSeries`).

    Returns:
        None
    """

    if not is_fitted:
        raise RuntimeError(
            "This forecaster is not fitted yet. Call `fit` with appropriate "
            "arguments before using `predict`."
        )

    if isinstance(steps, (int, np.integer)) and steps < 1:
        raise ValueError(
            f"`steps` must be an integer greater than or equal to 1. Got {steps}."
        )

    if isinstance(steps, list) and min(steps) < 1:
        raise ValueError(
            f"`steps` must be a list of integers greater than or equal to 1. Got {steps}."
        )

    if max_step is not None:
        if isinstance(steps, (int, np.integer)):
            if steps > max_step:
                raise ValueError(
                    f"The maximum step that can be predicted is {max_step}. "
                    f"Got {steps}."
                )
        elif isinstance(steps, list):
            if max(steps) > max_step:
                raise ValueError(
                    f"The maximum step that can be predicted is {max_step}. "
                    f"Got {max(steps)}."
                )

    if interval is not None or alpha is not None:
        check_interval(interval=interval, alpha=alpha)

    if exog_in_ and exog is None:
        raise ValueError(
            "Forecaster trained with exogenous variable/s. "
            "Same variable/s must be provided when predicting."
        )

    if not exog_in_ and exog is not None:
        raise ValueError(
            "Forecaster trained without exogenous variable/s. "
            "`exog` must be `None` when predicting."
        )

    if exog is not None:
        # If exog is a dictionary, it is assumed that it contains the exogenous
        # variables for each series.
        if isinstance(exog, dict):
            # Check that all series have the exogenous variables
            if levels is None and series_names_in_ is not None:
                levels = series_names_in_

            if isinstance(levels, str):
                levels = [levels]

            if levels is not None:
                for level in levels:
                    if level not in exog:
                        raise ValueError(
                            f"Exogenous variables for series '{level}' are missing."
                        )
                    check_exog(
                        exog=exog[level],
                        allow_nan=False,
                        series_id=f"`exog` for series '{level}'",
                    )
                    check_exog_dtypes(
                        exog=exog[level],
                        call_check_exog=False,
                        series_id=f"`exog` for series '{level}'",
                    )

                    # Check that exogenous variables are the same as used in training
                    # Get the name of columns
                    if isinstance(exog[level], pd.Series):
                        exog_names = [exog[level].name]
                    else:
                        exog_names = exog[level].columns.tolist()

                    if len(set(exog_names) - set(exog_names_in_)) > 0:
                        raise ValueError(
                            f"Exogenous variables must be: {exog_names_in_}. "
                            f"Got {exog_names} for series '{level}'."
                        )
        else:
            check_exog(exog=exog, allow_nan=False)
            check_exog_dtypes(exog=exog, call_check_exog=False)

            # Check that exogenous variables are the same as used in training
            # Get the name of columns
            if isinstance(exog, pd.Series):
                exog_names = [exog.name]
            else:
                exog_names = exog.columns.tolist()

            if len(set(exog_names) - set(exog_names_in_)) > 0:
                raise ValueError(
                    f"Exogenous variables must be: {exog_names_in_}. Got {exog_names}."
                )

    # Check last_window
    if last_window is not None:
        if isinstance(last_window, pd.DataFrame):
            if last_window.isna().to_numpy().any():
                raise ValueError("`last_window` has missing values.")
        else:
            check_y(last_window, series_id="`last_window`")

    return



"""
Base class for time series cross-validation splitting.
"""

from __future__ import annotations
import warnings
import numpy as np
import pandas as pd
from spotforecast2.exceptions import IgnoredArgumentWarning


class BaseFold:
    """
    Base class for all Fold classes in spotforecast. All fold classes should specify
    all the parameters that can be set at the class level in their ``__init__``.

    Args:
        steps (int, optional): Number of observations used to be predicted in each fold.
            This is also commonly referred to as the forecast horizon or test size.
            Defaults to None.
        initial_train_size (int | str | pd.Timestamp, optional): Number of observations
            used for initial training.

            - If an integer, the number of observations used for initial training.
            - If a date string or pandas Timestamp, it is the last date included in
              the initial training set.
            Defaults to None.
        fold_stride (int, optional): Number of observations that the start of the test
            set advances between consecutive folds.

            - If `None`, it defaults to the same value as `steps`, meaning that folds
              are placed back-to-back without overlap.
            - If `fold_stride < steps`, test sets overlap and multiple forecasts will
              be generated for the same observations.
            - If `fold_stride > steps`, gaps are left between consecutive test sets.
            Defaults to None.
        window_size (int, optional): Number of observations needed to generate the
            autoregressive predictors. Defaults to None.
        differentiation (int, optional): Number of observations to use for differentiation.
            This is used to extend the `last_window` as many observations as the
            differentiation order. Defaults to None.
        refit (bool | int, optional): Whether to refit the forecaster in each fold.

            - If `True`, the forecaster is refitted in each fold.
            - If `False`, the forecaster is trained only in the first fold.
            - If an integer, the forecaster is trained in the first fold and then refitted
              every `refit` folds.
            Defaults to False.
        fixed_train_size (bool, optional): Whether the training size is fixed or increases
            in each fold. Defaults to True.
        gap (int, optional): Number of observations between the end of the training set
            and the start of the test set. Defaults to 0.
        skip_folds (int | list, optional): Number of folds to skip.

            - If an integer, every 'skip_folds'-th is returned.
            - If a list, the indexes of the folds to skip.

            For example, if `skip_folds=3` and there are 10 folds, the returned folds are
            0, 3, 6, and 9. If `skip_folds=[1, 2, 3]`, the returned folds are 0, 4, 5, 6, 7,
            8, and 9. Defaults to None.
        allow_incomplete_fold (bool, optional): Whether to allow the last fold to include
            fewer observations than `steps`. If `False`, the last fold is excluded if it
            is incomplete. Defaults to True.
        return_all_indexes (bool, optional): Whether to return all indexes or only the
            start and end indexes of each fold. Defaults to False.
        verbose (bool, optional): Whether to print information about generated folds.
            Defaults to True.

    Attributes:
        initial_train_size (int): Number of observations used for initial training.
        window_size (int): Number of observations needed to generate the
            autoregressive predictors.
        differentiation (int): Number of observations to use for differentiation.
            This is used to extend the `last_window` as many observations as the
            differentiation order.
        return_all_indexes (bool): Whether to return all indexes or only the start
            and end indexes of each fold.
        verbose (bool): Whether to print information about generated folds.
    """

    def __init__(
        self,
        steps: int | None = None,
        initial_train_size: int | str | pd.Timestamp | None = None,
        fold_stride: int | None = None,
        window_size: int | None = None,
        differentiation: int | None = None,
        refit: bool | int = False,
        fixed_train_size: bool = True,
        gap: int = 0,
        skip_folds: int | list[int] | None = None,
        allow_incomplete_fold: bool = True,
        return_all_indexes: bool = False,
        verbose: bool = True,
    ) -> None:

        self._validate_params(
            cv_name=type(self).__name__,
            steps=steps,
            initial_train_size=initial_train_size,
            fold_stride=fold_stride,
            window_size=window_size,
            differentiation=differentiation,
            refit=refit,
            fixed_train_size=fixed_train_size,
            gap=gap,
            skip_folds=skip_folds,
            allow_incomplete_fold=allow_incomplete_fold,
            return_all_indexes=return_all_indexes,
            verbose=verbose,
        )

        self.initial_train_size = initial_train_size
        self.window_size = window_size
        self.differentiation = differentiation
        self.return_all_indexes = return_all_indexes
        self.verbose = verbose

    def _validate_params(
        self,
        cv_name: str,
        steps: int | None = None,
        initial_train_size: int | str | pd.Timestamp | None = None,
        fold_stride: int | None = None,
        window_size: int | None = None,
        differentiation: int | None = None,
        refit: bool | int = False,
        fixed_train_size: bool = True,
        gap: int = 0,
        skip_folds: int | list[int] | None = None,
        allow_incomplete_fold: bool = True,
        return_all_indexes: bool = False,
        verbose: bool = True,
        **kwargs,
    ) -> None:
        """
        Validate all input parameters to ensure correctness.
        """

        if cv_name == "TimeSeriesFold":
            if not isinstance(steps, (int, np.integer)) or steps < 1:
                raise ValueError(
                    f"`steps` must be an integer greater than 0. Got {steps}."
                )
            if not isinstance(
                initial_train_size, (int, np.integer, str, pd.Timestamp, type(None))
            ):
                raise ValueError(
                    f"`initial_train_size` must be an integer greater than 0, a date "
                    f"string, a pandas Timestamp, or None. Got {initial_train_size}."
                )
            if (
                isinstance(initial_train_size, (int, np.integer))
                and initial_train_size < 1
            ):
                raise ValueError(
                    f"`initial_train_size` must be an integer greater than 0, "
                    f"a date string, a pandas Timestamp, or None. Got {initial_train_size}."
                )
            if fold_stride is not None:
                if not isinstance(fold_stride, (int, np.integer)) or fold_stride < 1:
                    raise ValueError(
                        f"`fold_stride` must be an integer greater than 0. Got {fold_stride}."
                    )
            if not isinstance(refit, (bool, int, np.integer)):
                raise TypeError(
                    f"`refit` must be a boolean or an integer equal or greater than 0. "
                    f"Got {refit}."
                )
            if (
                isinstance(refit, (int, np.integer))
                and not isinstance(refit, bool)
                and refit < 0
            ):
                raise TypeError(
                    f"`refit` must be a boolean or an integer equal or greater than 0. "
                    f"Got {refit}."
                )
            if not isinstance(fixed_train_size, bool):
                raise TypeError(
                    f"`fixed_train_size` must be a boolean: `True`, `False`. "
                    f"Got {fixed_train_size}."
                )
            if not isinstance(gap, (int, np.integer)) or gap < 0:
                raise ValueError(
                    f"`gap` must be an integer greater than or equal to 0. Got {gap}."
                )
            if skip_folds is not None:
                if not isinstance(skip_folds, (int, np.integer, list, type(None))):
                    raise TypeError(
                        f"`skip_folds` must be an integer greater than 0, a list of "
                        f"integers or `None`. Got {skip_folds}."
                    )
                if isinstance(skip_folds, (int, np.integer)) and skip_folds < 1:
                    raise ValueError(
                        f"`skip_folds` must be an integer greater than 0, a list of "
                        f"integers or `None`. Got {skip_folds}."
                    )
                if isinstance(skip_folds, list) and any([x < 1 for x in skip_folds]):
                    raise ValueError(
                        f"`skip_folds` list must contain integers greater than or "
                        f"equal to 1. The first fold is always needed to train the "
                        f"forecaster. Got {skip_folds}."
                    )
            if not isinstance(allow_incomplete_fold, bool):
                raise TypeError(
                    f"`allow_incomplete_fold` must be a boolean: `True`, `False`. "
                    f"Got {allow_incomplete_fold}."
                )

        if cv_name == "OneStepAheadFold":
            if not isinstance(initial_train_size, (int, np.integer, str, pd.Timestamp)):
                raise ValueError(
                    f"`initial_train_size` must be an integer greater than 0, a date "
                    f"string, or a pandas Timestamp. Got {initial_train_size}."
                )
            if (
                isinstance(initial_train_size, (int, np.integer))
                and initial_train_size < 1
            ):
                raise ValueError(
                    f"`initial_train_size` must be an integer greater than 0, "
                    f"a date string, or a pandas Timestamp. Got {initial_train_size}."
                )

        if (
            not isinstance(window_size, (int, np.integer, pd.DateOffset, type(None)))
            or isinstance(window_size, (int, np.integer))
            and window_size < 1
        ):
            raise ValueError(
                f"`window_size` must be an integer greater than 0. Got {window_size}."
            )

        if differentiation is not None:
            if (
                not isinstance(differentiation, (int, np.integer))
                or differentiation < 0
            ):
                raise ValueError(
                    f"`differentiation` must be None or an integer greater than or "
                    f"equal to 0. Got {differentiation}."
                )

        if not isinstance(return_all_indexes, bool):
            raise TypeError(
                f"`return_all_indexes` must be a boolean: `True`, `False`. "
                f"Got {return_all_indexes}."
            )

        if not isinstance(verbose, bool):
            raise TypeError(
                f"`verbose` must be a boolean: `True`, `False`. " f"Got {verbose}."
            )

    def _extract_index(
        self,
        X: pd.Series | pd.DataFrame | pd.Index | dict[str, pd.Series | pd.DataFrame],
    ) -> pd.Index:
        """
        Extracts and returns the index from the input data X.

        Args:
            X (pd.Series | pd.DataFrame | pd.Index | dict): Time series data or
                index to split.

        Returns:
            pd.Index: Index extracted from the input data.
        """

        if isinstance(X, (pd.Series, pd.DataFrame)):
            idx = X.index
        elif isinstance(X, dict):
            indexes_freq = set()
            not_valid_index = []
            min_index = []
            max_index = []
            for k, v in X.items():
                if v is None:
                    continue

                idx = v.index
                if isinstance(idx, pd.DatetimeIndex):
                    indexes_freq.add(idx.freq)
                elif isinstance(idx, pd.RangeIndex):
                    indexes_freq.add(idx.step)
                else:
                    not_valid_index.append(k)

                min_index.append(idx[0])
                max_index.append(idx[-1])

            if not_valid_index:
                raise TypeError(
                    f"If `X` is a dictionary, all series must have a Pandas "
                    f"RangeIndex or DatetimeIndex with the same step/frequency. "
                    f"Review series: {not_valid_index}"
                )

            if None in indexes_freq:
                raise ValueError(
                    "If `X` is a dictionary, all series must have a Pandas "
                    "RangeIndex or DatetimeIndex with the same step/frequency. "
                    "Found series with no frequency or step."
                )
            if not len(indexes_freq) == 1:
                raise ValueError(
                    f"If `X` is a dictionary, all series must have a Pandas "
                    f"RangeIndex or DatetimeIndex with the same step/frequency. "
                    f"Found frequencies: {sorted(indexes_freq)}"
                )

            if isinstance(idx, pd.DatetimeIndex):
                idx = pd.date_range(
                    start=min(min_index), end=max(max_index), freq=indexes_freq.pop()
                )
            else:
                idx = pd.RangeIndex(
                    start=min(min_index),
                    stop=max(max_index) + 1,
                    step=indexes_freq.pop(),
                )
        else:
            idx = X

        return idx

    def set_params(self, params: dict) -> None:
        """
        Set the parameters of the Fold object. Before overwriting the current
        parameters, the input parameters are validated to ensure correctness.

        Args:
            params (dict): Dictionary with the parameters to set.
        """

        if not isinstance(params, dict):
            raise TypeError(f"`params` must be a dictionary. Got {type(params)}.")

        current_params = dict(vars(self))
        unknown_params = set(params.keys()) - set(current_params.keys())
        if unknown_params:
            warnings.warn(
                f"Unknown parameters: {unknown_params}. They have been ignored.",
                IgnoredArgumentWarning,
            )

        filtered_params = {k: v for k, v in params.items() if k in current_params}
        updated_params = {
            "cv_name": type(self).__name__,
            **current_params,
            **filtered_params,
        }

        self._validate_params(**updated_params)
        for key, value in updated_params.items():
            setattr(self, key, value)



"""
Data transformation utilities for time series forecasting.

This module provides functions for normalizing and transforming data formats.
"""

from typing import Union
import numpy as np
import pandas as pd


def input_to_frame(
    data: Union[pd.Series, pd.DataFrame], input_name: str
) -> pd.DataFrame:
    """
    Convert input data to a pandas DataFrame.

    This function ensures consistent DataFrame format for internal processing.
    If data is already a DataFrame, it's returned as-is. If it's a Series,
    it's converted to a single-column DataFrame.

    Args:
        data: Input data as pandas Series or DataFrame.
        input_name: Name of the input data type. Accepted values are:
            - 'y': Target time series
            - 'last_window': Last window for prediction
            - 'exog': Exogenous variables

    Returns:
        DataFrame version of the input data. For Series input, uses the series
        name if available, otherwise uses a default name based on input_name.

    Examples:
        >>> import pandas as pd
        >>> from spotforecast2.utils.data_transform import input_to_frame
        >>>
        >>> # Series with name
        >>> y = pd.Series([1, 2, 3], name="sales")
        >>> df = input_to_frame(y, input_name="y")
        >>> df.columns.tolist()
        ['sales']
        >>>
        >>> # Series without name (uses default)
        >>> y_no_name = pd.Series([1, 2, 3])
        >>> df = input_to_frame(y_no_name, input_name="y")
        >>> df.columns.tolist()
        ['y']
        >>>
        >>> # DataFrame (returned as-is)
        >>> df_input = pd.DataFrame({"temp": [20, 21], "humidity": [50, 55]})
        >>> df_output = input_to_frame(df_input, input_name="exog")
        >>> df_output.columns.tolist()
        ['temp', 'humidity']
        >>>
        >>> # Exog series without name
        >>> exog = pd.Series([10, 20, 30])
        >>> df_exog = input_to_frame(exog, input_name="exog")
        >>> df_exog.columns.tolist()
        ['exog']
    """
    output_col_name = {"y": "y", "last_window": "y", "exog": "exog"}

    if isinstance(data, pd.Series):
        data = data.to_frame(
            name=data.name if data.name is not None else output_col_name[input_name]
        )

    return data


def expand_index(index: Union[pd.Index, None], steps: int) -> pd.Index:
    """
    Create a new index extending from the end of the original index.

    This function generates future indices for forecasting by extending the time
    series index by a specified number of steps. Handles both DatetimeIndex and
    RangeIndex appropriately.

    Args:
        index: Original pandas Index (DatetimeIndex or RangeIndex). If None,
            creates a RangeIndex starting from 0.
        steps: Number of future steps to generate.

    Returns:
        New pandas Index with `steps` future periods.

    Raises:
        TypeError: If steps is not an integer, or if index is neither DatetimeIndex
            nor RangeIndex.

    Examples:
        >>> import pandas as pd
        >>> from spotforecast2.utils.data_transform import expand_index
        >>>
        >>> # DatetimeIndex
        >>> dates = pd.date_range("2023-01-01", periods=5, freq="D")
        >>> new_index = expand_index(dates, 3)
        >>> new_index
        DatetimeIndex(['2023-01-06', '2023-01-07', '2023-01-08'], dtype='datetime64[ns]', freq='D')
        >>>
        >>> # RangeIndex
        >>> range_idx = pd.RangeIndex(start=0, stop=10)
        >>> new_index = expand_index(range_idx, 5)
        >>> new_index
        RangeIndex(start=10, stop=15, step=1)
        >>>
        >>> # None index (creates new RangeIndex)
        >>> new_index = expand_index(None, 3)
        >>> new_index
        RangeIndex(start=0, stop=3, step=1)
        >>>
        >>> # Invalid: steps not an integer
        >>> try:
        ...     expand_index(dates, 3.5)
        ... except TypeError as e:
        ...     print("Error: steps must be an integer")
        Error: steps must be an integer
    """
    if not isinstance(steps, (int, np.integer)):
        raise TypeError(f"`steps` must be an integer. Got {type(steps)}.")

    # Convert numpy integer to Python int if needed
    if isinstance(steps, np.integer):
        steps = int(steps)

    if isinstance(index, pd.Index):
        if isinstance(index, pd.DatetimeIndex):
            new_index = pd.date_range(
                start=index[-1] + index.freq, periods=steps, freq=index.freq
            )
        elif isinstance(index, pd.RangeIndex):
            new_index = pd.RangeIndex(start=index[-1] + 1, stop=index[-1] + 1 + steps)
        else:
            raise TypeError(
                "Argument `index` must be a pandas DatetimeIndex or RangeIndex."
            )
    else:
        new_index = pd.RangeIndex(start=0, stop=steps)

    return new_index


def transform_dataframe(
    df: pd.DataFrame,
    transformer: object,
    fit: bool = False,
    inverse_transform: bool = False,
) -> pd.DataFrame:
    """
    Transform raw values of pandas DataFrame with a scikit-learn alike
    transformer, preprocessor or ColumnTransformer.

    The transformer used must have the following methods: fit, transform,
    fit_transform and inverse_transform. ColumnTransformers are not allowed
    since they do not have inverse_transform method.

    Args:
        df: DataFrame to be transformed.
        transformer: Scikit-learn alike transformer, preprocessor, or ColumnTransformer.
            Must implement fit, transform, fit_transform and inverse_transform.
        fit: Train the transformer before applying it. Defaults to False.
        inverse_transform: Transform back the data to the original representation.
            This is not available when using transformers of class
            scikit-learn ColumnTransformers. Defaults to False.

    Returns:
        Transformed DataFrame.

    Raises:
        TypeError: If df is not a pandas DataFrame.
        ValueError: If inverse_transform is requested for ColumnTransformer.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"`df` argument must be a pandas DataFrame. Got {type(df)}")

    if transformer is None:
        return df

    # Check for ColumnTransformer by class name to avoid importing sklearn
    is_column_transformer = type(
        transformer
    ).__name__ == "ColumnTransformer" or hasattr(transformer, "transformers")

    if inverse_transform and is_column_transformer:
        raise ValueError(
            "`inverse_transform` is not available when using ColumnTransformers."
        )

    if not inverse_transform:
        if fit:
            values_transformed = transformer.fit_transform(df)
        else:
            values_transformed = transformer.transform(df)
    else:
        values_transformed = transformer.inverse_transform(df)

    if hasattr(values_transformed, "toarray"):
        # If the returned values are in sparse matrix format, it is converted to dense
        values_transformed = values_transformed.toarray()

    if isinstance(values_transformed, pd.DataFrame):
        df_transformed = values_transformed
    else:
        df_transformed = pd.DataFrame(
            values_transformed, index=df.index, columns=df.columns
        )

    return df_transformed



"""
QuantileBinner class for binning data into quantile-based bins.

This module contains the QuantileBinner class which bins data into quantile-based bins
using numpy.percentile with optimized performance using numpy.searchsorted.
"""

from __future__ import annotations
import warnings
import numpy as np
from typing import Any
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.exceptions import NotFittedError

from spotforecast2.exceptions import IgnoredArgumentWarning


class QuantileBinner(BaseEstimator, TransformerMixin):
    """
    Bin data into quantile-based bins using numpy.percentile.

    This class is similar to sklearn's KBinsDiscretizer but optimized for
    performance using numpy.searchsorted for fast bin assignment. Bin intervals
    are defined following the convention: bins[i-1] <= x < bins[i]. Values
    outside the range are clipped to the first or last bin.

    Args:
        n_bins: The number of quantile-based bins to create. Must be >= 2.
        method: The method used to compute quantiles, passed to numpy.percentile.
            Default is 'linear'. Valid values: "inverse_cdf",
            "averaged_inverse_cdf", "closest_observation",
            "interpolated_inverse_cdf", "hazen", "weibull", "linear",
            "median_unbiased", "normal_unbiased".
        subsample: Maximum number of samples for computing quantiles. If dataset
            has more samples, a random subset is used. Default 200000.
        dtype: Data type for bin indices. Default is numpy.float64.
        random_state: Random seed for subset generation. Default 789654.

    Attributes:
        n_bins (int): Number of bins to create.
        method (str): Quantile computation method.
        subsample (int): Maximum samples for quantile computation.
        dtype (type): Data type for bin indices.
        random_state (int): Random seed.
        n_bins_ (int): Actual number of bins after fitting (may differ from n_bins
            if duplicate edges are found).
        bin_edges_ (np.ndarray): Edges of the bins learned during fitting.
        internal_edges_ (np.ndarray): Internal edges for optimized bin assignment.
        intervals_ (dict): Mapping from bin index to (lower, upper) interval bounds.

    Examples:
        >>> import numpy as np
        >>> from spotforecast2.preprocessing import QuantileBinner
        >>>
        >>> # Basic usage: create 3 quantile bins
        >>> X = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        >>> binner = QuantileBinner(n_bins=3)
        >>> _ = binner.fit(X)
        >>> result = binner.transform(np.array([1.5, 5.5, 9.5]))
        >>> print(result)
        [0. 1. 2.]
        >>>
        >>> # Check bin intervals
        >>> print(binner.n_bins_)
        3
        >>> assert len(binner.intervals_) == 3
        >>>
        >>> # Use fit_transform for one-step operation
        >>> X2 = np.array([10, 20, 30, 40, 50])
        >>> binner2 = QuantileBinner(n_bins=2)
        >>> bins = binner2.fit_transform(X2)
        >>> print(bins)
        [0. 0. 1. 1. 1.]
    """

    def __init__(
        self,
        n_bins: int,
        method: str = "linear",
        subsample: int = 200000,
        dtype: type = np.float64,
        random_state: int = 789654,
    ) -> None:

        self._validate_params(n_bins, method, subsample, dtype, random_state)

        self.n_bins = n_bins
        self.method = method
        self.subsample = subsample
        self.dtype = dtype
        self.random_state = random_state
        self.n_bins_ = None
        self.bin_edges_ = None
        self.internal_edges_ = None
        self.intervals_ = None

    def _validate_params(
        self, n_bins: int, method: str, subsample: int, dtype: type, random_state: int
    ):
        """
        Validate parameters passed to the class initializer.

        Args:
            n_bins: Number of quantile-based bins. Must be int >= 2.
            method: Quantile computation method for numpy.percentile.
            subsample: Number of samples for computing quantiles. Must be int >= 1.
            dtype: Data type for bin indices. Must be a valid numpy dtype.
            random_state: Random seed for subset generation. Must be int >= 0.

        Raises:
            ValueError: If n_bins < 2, method is invalid, subsample < 1,
                random_state < 0, or dtype is not a valid type.

        Examples:
            >>> import numpy as np
            >>> from spotforecast2.preprocessing import QuantileBinner
            >>>
            >>> # Valid parameters work fine
            >>> binner = QuantileBinner(n_bins=5, method='linear')
            >>> assert binner.n_bins == 5
            >>>
            >>> # Invalid n_bins raises ValueError
            >>> try:
            ...     binner = QuantileBinner(n_bins=1)
            ... except ValueError as e:
            ...     assert 'greater than 1' in str(e)
            >>>
            >>> # Invalid method raises ValueError
            >>> try:
            ...     binner = QuantileBinner(n_bins=3, method='invalid')
            ... except ValueError as e:
            ...     assert 'must be one of' in str(e)
        """

        if not isinstance(n_bins, int) or n_bins < 2:
            raise ValueError(f"`n_bins` must be an int greater than 1. Got {n_bins}.")

        valid_methods = [
            "inverse_cdf",
            "averaged_inverse_cdf",
            "closest_observation",
            "interpolated_inverse_cdf",
            "hazen",
            "weibull",
            "linear",
            "median_unbiased",
            "normal_unbiased",
        ]
        if method not in valid_methods:
            raise ValueError(f"`method` must be one of {valid_methods}. Got {method}.")
        if not isinstance(subsample, int) or subsample < 1:
            raise ValueError(
                f"`subsample` must be an integer greater than or equal to 1. "
                f"Got {subsample}."
            )
        if not isinstance(random_state, int) or random_state < 0:
            raise ValueError(
                f"`random_state` must be an integer greater than or equal to 0. "
                f"Got {random_state}."
            )
        if not isinstance(dtype, type):
            raise ValueError(f"`dtype` must be a valid numpy dtype. Got {dtype}.")

    def fit(self, X: np.ndarray, y: object = None) -> object:
        """
        Learn bin edges based on quantiles from training data.

        Computes quantile-based bin edges using numpy.percentile. If the dataset
        contains more samples than `subsample`, a random subset is used. Duplicate
        edges (which can occur with repeated values) are removed automatically.

        Args:
            X: Training data (1D numpy array) for computing quantiles.
            y: Ignored.

        Returns:
            Self for method chaining.

        Raises:
            ValueError: If input data X is empty.

        Examples:
            >>> import numpy as np
            >>> from spotforecast2.preprocessing import QuantileBinner
            >>>
            >>> # Fit with basic data
            >>> X = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
            >>> binner = QuantileBinner(n_bins=3)
            >>> _ = binner.fit(X)
            >>> print(binner.n_bins_)
            3
            >>> print(len(binner.bin_edges_))
            4
            >>>
            >>> # Fit with repeated values (may reduce number of bins)
            >>> X_repeated = np.array([1, 1, 1, 2, 2, 2, 3, 3, 3])
            >>> binner2 = QuantileBinner(n_bins=5)
            >>> _ = binner2.fit(X_repeated)
            >>> # n_bins_ may be less than 5 due to duplicates
            >>> assert binner2.n_bins_ <= 5
        """
        # Note: Original implementation expects X, but sklearn TransformerMixin passes y=None.
        # Adjusted signature to (self, X: np.ndarray, y: object = None)

        if X.size == 0:
            raise ValueError("Input data `X` cannot be empty.")
        if len(X) > self.subsample:
            rng = np.random.default_rng(self.random_state)
            X = X[rng.integers(0, len(X), self.subsample)]

        bin_edges = np.percentile(
            a=X, q=np.linspace(0, 100, self.n_bins + 1), method=self.method
        )

        # Remove duplicate edges (can happen when data has many repeated values)
        # to ensure bins are always numbered 0 to n_bins_-1
        self.bin_edges_ = np.unique(bin_edges)

        # Ensure at least 1 bin when all values are identical
        if len(self.bin_edges_) == 1:
            # Create artificial edges around the single value
            self.bin_edges_ = np.array([self.bin_edges_.item(), self.bin_edges_.item()])

        self.n_bins_ = len(self.bin_edges_) - 1

        if self.n_bins_ != self.n_bins:
            warnings.warn(
                f"The number of bins has been reduced from {self.n_bins} to "
                f"{self.n_bins_} due to duplicated edges caused by repeated predicted "
                f"values.",
                IgnoredArgumentWarning,
            )

        # Internal edges for optimized transform with searchsorted
        self.internal_edges_ = self.bin_edges_[1:-1]
        self.intervals_ = {
            int(i): (float(self.bin_edges_[i]), float(self.bin_edges_[i + 1]))
            for i in range(self.n_bins_)
        }

        return self

    def transform(self, X: np.ndarray, y: object = None) -> np.ndarray:
        """
        Assign new data to learned bins.

        Uses numpy.searchsorted for efficient bin assignment. Values are assigned
        to bins following the convention: bins[i-1] <= x < bins[i]. Values outside
        the fitted range are clipped to the first or last bin.

        Args:
            X: Data to assign to bins (1D numpy array).
            y: Ignored.

        Returns:
            Bin indices as numpy array with dtype specified in __init__.

        Raises:
            NotFittedError: If fit() has not been called yet.

        Examples:
            >>> import numpy as np
            >>> from spotforecast2.preprocessing import QuantileBinner
            >>>
            >>> # Fit and transform
            >>> X_train = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
            >>> binner = QuantileBinner(n_bins=3)
            >>> _ = binner.fit(X_train)
            >>>
            >>> X_test = np.array([1.5, 5.5, 9.5])
            >>> result = binner.transform(X_test)
            >>> print(result)
            [0. 1. 2.]
            >>>
            >>> # Values outside range are clipped
            >>> X_extreme = np.array([0, 100])
            >>> result_extreme = binner.transform(X_extreme)
            >>> print(result_extreme)  # Both clipped to valid bin indices
            [0. 2.]
        """

        if self.bin_edges_ is None:
            raise NotFittedError(
                "The model has not been fitted yet. Call 'fit' with training data first."
            )

        bin_indices = np.searchsorted(self.internal_edges_, X, side="right").astype(
            self.dtype
        )

        return bin_indices

    def fit_transform(self, X, y=None, **fit_params):
        """
        Fit to data, then transform it.

        Fits transformer to X and y with optional parameters fit_params
        and returns a transformed version of X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Input samples.

        y :  array-like of shape (n_samples,) or (n_samples, n_outputs), \
                default=None
            Target values (None for unsupervised transformations).

        **fit_params : dict
            Additional fit parameters.

        Returns
        -------
        X_new : ndarray array of shape (n_samples, n_features_new)
            Transformed array.
        """
        # fit_transform is usually provided by TransformerMixin but we can implement it
        # or rely on inheritance. The original implementation had it explicitly.

        self.fit(X, y)
        return self.transform(X, y)

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """
        Get parameters of the quantile binner.

        Returns:
            Dictionary containing n_bins, method, subsample, dtype, and
            random_state parameters.

        Examples:
            >>> import numpy as np
            >>> from spotforecast2.preprocessing import QuantileBinner
            >>>
            >>> binner = QuantileBinner(n_bins=5, method='median_unbiased', subsample=1000)
            >>> params = binner.get_params()
            >>> print(params['n_bins'])
            5
            >>> print(params['method'])
            median_unbiased
            >>> print(params['subsample'])
            1000
        """

        return {
            "n_bins": self.n_bins,
            "method": self.method,
            "subsample": self.subsample,
            "dtype": self.dtype,
            "random_state": self.random_state,
        }

    def set_params(self, **params: Any) -> "QuantileBinner":
        """
        Set parameters of the QuantileBinner.

        Args:
            **params: Parameter names and values to set as keyword arguments.

        Returns:
            self: Returns the updated QuantileBinner instance.

        Examples:
            >>> import numpy as np
            >>> from spotforecast2.preprocessing import QuantileBinner
            >>>
            >>> binner = QuantileBinner(n_bins=3)
            >>> print(binner.n_bins)
            3
            >>> binner.set_params(n_bins=5, method='weibull')
            >>> print(binner.n_bins)
            5
            >>> print(binner.method)
            weibull
        """

        for param, value in params.items():
            setattr(self, param, value)
        return self



"""Weather data fetching and processing using Open-Meteo API."""

import logging
from pathlib import Path
from typing import Optional, Dict, Any, Union

import pandas as pd
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


class WeatherClient:
    """Client for fetching weather data from Open-Meteo API.

    Handles the low-level API interactions, parameter building, and response parsing.
    """

    ARCHIVE_BASE_URL = "https://archive-api.open-meteo.com/v1/archive"
    FORECAST_BASE_URL = "https://api.open-meteo.com/v1/forecast"

    HOURLY_PARAMS = [
        "temperature_2m",
        "relative_humidity_2m",
        "precipitation",
        "rain",
        "snowfall",
        "weather_code",
        "pressure_msl",
        "surface_pressure",
        "cloud_cover",
        "cloud_cover_low",
        "cloud_cover_mid",
        "cloud_cover_high",
        "wind_speed_10m",
        "wind_direction_10m",
        "wind_gusts_10m",
    ]

    def __init__(self, latitude: float, longitude: float):
        """Initialize WeatherClient.

        Args:
            latitude: Latitude of the location.
            longitude: Longitude of the location.
        """
        self.latitude = latitude
        self.longitude = longitude
        self.logger = logging.getLogger(__name__)
        self._session = self._create_session()

    def _create_session(self) -> requests.Session:
        """Create a requests session with retry logic."""
        session = requests.Session()
        retry_strategy = Retry(
            total=3,
            backoff_factor=1,
            status_forcelist=[429, 500, 502, 503, 504],
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session.mount("https://", adapter)
        session.mount("http://", adapter)
        return session

    def _fetch(self, url: str, params: Dict[str, Any]) -> pd.DataFrame:
        """Execute API request and return parsed DataFrame."""
        try:
            response = self._session.get(url, params=params, timeout=30)
            response.raise_for_status()
            data = response.json()
        except requests.exceptions.RequestException as e:
            self.logger.error(f"API request failed: {e}")
            raise

        if "error" in data and data["error"]:
            raise ValueError(
                f"Open-Meteo API error: {data.get('reason', 'Unknown error')}"
            )

        hourly_data = data.get("hourly", {})
        if not hourly_data:
            raise ValueError("No hourly data returned from API")

        # Parse to DataFrame
        times = pd.to_datetime(hourly_data["time"])
        df_dict = {"datetime": times}
        for param in self.HOURLY_PARAMS:
            if param in hourly_data:
                df_dict[param] = hourly_data[param]

        df = pd.DataFrame(df_dict)
        df.set_index("datetime", inplace=True)
        return df

    def fetch_archive(
        self, start: pd.Timestamp, end: pd.Timestamp, timezone: str = "UTC"
    ) -> pd.DataFrame:
        """Fetch historical data from Archive API."""
        params = {
            "latitude": self.latitude,
            "longitude": self.longitude,
            "hourly": ",".join(self.HOURLY_PARAMS),
            "timezone": timezone,
            "start_date": start.strftime("%Y-%m-%d"),
            "end_date": end.strftime("%Y-%m-%d"),
        }
        return self._fetch(self.ARCHIVE_BASE_URL, params)

    def fetch_forecast(self, days_ahead: int, timezone: str = "UTC") -> pd.DataFrame:
        """Fetch forecast data from Forecast API."""
        params = {
            "latitude": self.latitude,
            "longitude": self.longitude,
            "hourly": ",".join(self.HOURLY_PARAMS),
            "timezone": timezone,
            "forecast_days": days_ahead,
        }
        return self._fetch(self.FORECAST_BASE_URL, params)


class WeatherService(WeatherClient):
    """High-level service for weather data generation.

    Extends WeatherClient with caching, hybrid fetching (archive+forecast),
    and fallback strategies.
    """

    def __init__(
        self,
        latitude: float,
        longitude: float,
        cache_path: Optional[Path] = None,
        use_forecast: bool = True,
    ):
        super().__init__(latitude, longitude)
        self.cache_path = cache_path
        self.use_forecast = use_forecast

    def get_dataframe(
        self,
        start: Union[str, pd.Timestamp],
        end: Union[str, pd.Timestamp],
        timezone: str = "UTC",
        freq: str = "h",
        fallback_on_failure: bool = True,
    ) -> pd.DataFrame:
        """Get weather DataFrame for a specified range using best available methods.

        Refactored from spotpredict.create_weather_df.
        """
        start_ts = pd.Timestamp(start)
        end_ts = pd.Timestamp(end)

        # Localize if naive
        if start_ts.tz is None:
            start_ts = start_ts.tz_localize(timezone)
        if end_ts.tz is None:
            end_ts = end_ts.tz_localize(timezone)

        # Convert to UTC for consistency
        start_utc = start_ts.tz_convert("UTC")
        end_utc = end_ts.tz_convert("UTC")

        # 1. Try Cache
        cached_df = self._load_cache()
        if cached_df is not None:
            if cached_df.index.min() <= start_utc and cached_df.index.max() >= end_utc:
                self.logger.info("Using full cached data.")
                return self._finalize_df(
                    cached_df.loc[start_utc:end_utc], freq, timezone
                )

        # 2. Hybrid Fetch (filling gaps if cache exists, or fetching all)
        # (The original logic did partial fills, but full fetch is safer and
        # simpler for now unless specifically improved).
        # Actually, strict refactor implies keeping logic. Let's keep it simple:
        # fetch what's needed.

        try:
            df = self._fetch_hybrid(start_ts, end_ts, timezone)
        except Exception as e:
            self.logger.warning(f"Fetch failed: {e}")
            if fallback_on_failure and cached_df is not None and len(cached_df) >= 24:
                df = self._create_fallback(start_utc, end_utc, cached_df, timezone)
            else:
                raise

        # 3. Merge with cache and save
        if cached_df is not None:
            df = pd.concat([cached_df, df])
            df = df[~df.index.duplicated(keep="last")].sort_index()  # Keep new data

        if self.cache_path:
            self._save_cache(df)

        # 4. Return slice
        return self._finalize_df(df.loc[start_utc:end_utc], freq, timezone)

    def _fetch_hybrid(
        self, start: pd.Timestamp, end: pd.Timestamp, timezone: str
    ) -> pd.DataFrame:
        """Fetch from Archive and/or Forecast based on date."""
        now = pd.Timestamp.now(tz=start.tz)
        archive_cutoff = now - pd.Timedelta(days=5)

        dfs = []

        # Archive part
        if start < archive_cutoff:
            arch_end = min(end, archive_cutoff)
            try:
                dfs.append(self.fetch_archive(start, arch_end, timezone))
            except Exception as e:
                self.logger.warning(f"Archive fetch warning: {e}")

        # Forecast part
        if end > now and self.use_forecast:
            days = (end - now).days + 2
            days = min(max(1, days), 16)
            try:
                df_fore = self.fetch_forecast(days, timezone)
                # Filter forecast to needed range to avoid overlap issues
                dfs.append(df_fore)
            except Exception as e:
                self.logger.warning(f"Forecast fetch warning: {e}")

        if not dfs:
            raise ValueError("Could not fetch data from Archive or Forecast.")

        full_df = pd.concat(dfs)
        full_df = full_df[~full_df.index.duplicated(keep="first")].sort_index()

        # Ensure UTC index
        if full_df.index.tz is None:
            full_df.index = full_df.index.tz_localize(timezone)
        full_df.index = full_df.index.tz_convert("UTC")

        return full_df

    def _create_fallback(
        self,
        start: pd.Timestamp,
        end: pd.Timestamp,
        source_df: pd.DataFrame,
        timezone: str,
    ) -> pd.DataFrame:
        """Repeat last 24h of data."""
        last_24 = source_df.tail(24)
        hours = int((end - start).total_seconds() / 3600) + 1
        repeats = (hours // 24) + 1

        new_data = pd.concat([last_24] * repeats, ignore_index=True)
        new_data = new_data.iloc[:hours]

        idx = pd.date_range(start, periods=hours, freq="h", tz="UTC")
        new_data.index = idx
        return new_data

    def _load_cache(self) -> Optional[pd.DataFrame]:
        if not self.cache_path or not self.cache_path.exists():
            return None
        try:
            df = pd.read_parquet(self.cache_path)
            if df.index.tz is None:
                df.index = df.index.tz_localize("UTC")
            return df
        except Exception:
            return None

    def _save_cache(self, df: pd.DataFrame):
        if self.cache_path:
            self.cache_path.parent.mkdir(parents=True, exist_ok=True)
            df.to_parquet(self.cache_path)

    def _finalize_df(self, df: pd.DataFrame, freq: str, timezone: str) -> pd.DataFrame:
        """Resample and localize."""
        # Resample
        if freq != "h":  # Assuming API returns hourly
            df = df.resample(freq).ffill()  # Forward fill for weather is reasonable

        # Fill gaps
        df = df.ffill().bfill()

        # Convert to requested timezone if needed (though we keep internal UTC mostly)
        # User requested specific tz output usually?
        # Original code returned normalized DF. Let's ensure frequency matches exactly.

        return df



"""Processing module for end-to-end forecasting pipelines."""

from .n2n_predict import n2n_predict
from .n2n_predict_with_covariates import n2n_predict_with_covariates

__all__ = [
    "n2n_predict",
    "n2n_predict_with_covariates",
]



from .split_ts_cv import TimeSeriesFold
from .split_one_step import OneStepAheadFold
from .validation import backtesting_forecaster

__all__ = ["TimeSeriesFold", "OneStepAheadFold", "backtesting_forecaster"]



from .fetch_data import (
    get_data_home,
    get_cache_home,
    fetch_data,
    fetch_holiday_data,
    fetch_weather_data,
)

__all__ = [
    "get_data_home",
    "get_cache_home",
    "fetch_data",
    "fetch_holiday_data",
    "fetch_weather_data",
]



from __future__ import annotations
import os
import numpy as np
import warnings
from typing import Callable
from copy import deepcopy
import pandas as pd
from joblib import cpu_count
from tqdm.auto import tqdm
from sklearn.model_selection import ParameterGrid

from spotforecast2.exceptions import (
    IgnoredArgumentWarning,
)
from spotforecast2.model_selection.split_ts_cv import TimeSeriesFold
from spotforecast2.model_selection.split_one_step import OneStepAheadFold
from spotforecast2.model_selection.utils_common import (
    initialize_lags_grid,
    check_backtesting_input,
    check_one_step_ahead_input,
    select_n_jobs_backtesting,
)
from spotforecast2.forecaster.metrics import add_y_train_argument, _get_metric
from spotforecast2.model_selection.utils_metrics import (
    _calculate_metrics_one_step_ahead,
)
from spotforecast2.model_selection.validation import _backtesting_forecaster
from spotforecast2.forecaster.utils import set_skforecast_warnings


def _evaluate_grid_hyperparameters(
    forecaster: object,
    y: pd.Series,
    cv: TimeSeriesFold | OneStepAheadFold,
    param_grid: dict[str, object],
    metric: str | Callable | list[str | Callable],
    exog: pd.Series | pd.DataFrame | None = None,
    lags_grid: (
        list[int | list[int] | np.ndarray[int] | range[int]]
        | dict[str, list[int | list[int] | np.ndarray[int] | range[int]]]
        | None
    ) = None,
    return_best: bool = True,
    n_jobs: int | str = "auto",
    verbose: bool = False,
    show_progress: bool = True,
    suppress_warnings: bool = False,
    output_file: str | None = None,
) -> pd.DataFrame:
    """
    Evaluate combinations of hyperparameters and lags for a given forecaster.
    """

    set_skforecast_warnings(suppress_warnings, action="ignore")

    forecaster = deepcopy(forecaster)
    forecaster_search = forecaster  # Alias for consistency with original code
    is_regression = (
        forecaster_search.__spotforecast_tags__["forecaster_task"] == "regression"
    )

    if isinstance(cv, TimeSeriesFold):
        check_backtesting_input(
            forecaster=forecaster,
            cv=cv,
            y=y,
            metric=metric,
            exog=exog,
            n_jobs=n_jobs,
            show_progress=show_progress,
            suppress_warnings=suppress_warnings,
        )
    else:
        # OneStepAheadFold
        check_one_step_ahead_input(
            forecaster=forecaster,
            cv=cv,
            y=y,
            metric=metric,
            exog=exog,
            show_progress=show_progress,
            suppress_warnings=suppress_warnings,
        )
        # Update cv params in case they were modified during input check or need setting
        # (Original code does initialization of initial_train_size here)
        # We assume cv is already correctly set up or updated by user.
        # But OneStepAheadFold in original is used to split?
        # Original code re-initializes cv params?
        # Lines 280-293 in original handle date_to_index_position for initial_train_size.
        # We should probably do that if passing strings.
        # But TimeSeriesFold does it in its init?
        # OneStepAheadFold might support string initial_train_size.
        # Let's adding it for robustness if needed, but keeping it simple for now as per prior porting.
        pass

    if not isinstance(metric, list):
        metric = [metric]
    metric = [
        _get_metric(metric=m) if isinstance(m, str) else add_y_train_argument(m)
        for m in metric
    ]
    metric_dict = {(m if isinstance(m, str) else m.__name__): [] for m in metric}

    if len(metric_dict) != len(metric):
        raise ValueError("When `metric` is a `list`, each metric name must be unique.")

    lags_grid, lags_label = initialize_lags_grid(forecaster, lags_grid)
    cv = deepcopy(cv)

    if n_jobs == "auto":
        refit = cv.refit if isinstance(cv, TimeSeriesFold) else False
        n_jobs = select_n_jobs_backtesting(forecaster=forecaster, refit=refit)
    elif isinstance(cv, TimeSeriesFold) and cv.refit != 1 and n_jobs != 1:
        warnings.warn(
            "If `refit` is an integer other than 1 (intermittent refit). `n_jobs` "
            "is set to 1 to avoid unexpected results during parallelization.",
            IgnoredArgumentWarning,
        )
        n_jobs = 1
    else:
        n_jobs = n_jobs if n_jobs > 0 else cpu_count()

    print(
        f"Number of models compared: {len(param_grid) * len(lags_grid)}. "
        f"Training models..."
    )

    if show_progress:
        lags_grid_tqdm = tqdm(lags_grid.items(), desc="Lags grid", position=0)
    else:
        lags_grid_tqdm = lags_grid.items()

    if output_file is not None and os.path.isfile(output_file):
        os.remove(output_file)

    lags_list = []
    lags_label_list = []
    params_list = []

    for lags_k, lags_v in lags_grid_tqdm:

        forecaster_search.set_lags(lags_v)
        lags_v = forecaster_search.lags.copy()
        if lags_label == "values":
            lags_k = lags_v

        # OneStepAhead split is done once per lag config if independent of params
        # But params might affect transformation?
        # In original code, split is done inside the loop over lags, before params loop.
        if isinstance(cv, OneStepAheadFold):
            X_train, y_train, X_test, y_test = (
                forecaster_search._train_test_split_one_step_ahead(
                    y=y, initial_train_size=cv.initial_train_size, exog=exog
                )
            )

        if show_progress:
            param_grid_tqdm = tqdm(
                param_grid, desc="Parameters grid", position=1, leave=False
            )
        else:
            param_grid_tqdm = param_grid

        for params in param_grid_tqdm:
            try:
                forecaster_search.set_params(**params)

                if isinstance(cv, TimeSeriesFold):
                    metric_values = _backtesting_forecaster(
                        forecaster=forecaster_search,
                        y=y,
                        cv=cv,
                        metric=metric,
                        exog=exog,
                        n_jobs=n_jobs,
                        verbose=verbose,
                        show_progress=False,
                        suppress_warnings=suppress_warnings,
                    )[0]
                    # metric_values is a DataFrame, we want list of values for the row (0)
                    metric_values = metric_values.iloc[0, :].to_list()
                else:
                    # One Step Ahead
                    metric_values = _calculate_metrics_one_step_ahead(
                        forecaster=forecaster_search,
                        metrics=metric,
                        X_train=X_train,
                        y_train=y_train,
                        X_test=X_test,
                        y_test=y_test,
                    )
            except Exception as e:
                warnings.warn(f"Parameters skipped: {params}. {e}", RuntimeWarning)
                continue

            # Filter warnings if needed/configured
            warnings.filterwarnings(
                "ignore",
                category=RuntimeWarning,
                message="The forecaster will be fit.*",
            )

            lags_list.append(lags_v)
            lags_label_list.append(lags_k)
            params_list.append(params)
            for m, m_value in zip(metric, metric_values):
                m_name = m if isinstance(m, str) else m.__name__
                metric_dict[m_name].append(m_value)

            if output_file is not None:
                header = [
                    "lags",
                    "lags_label",
                    "params",
                    *metric_dict.keys(),
                    *params.keys(),
                ]
                row = [lags_v, lags_k, params, *metric_values, *params.values()]
                if not os.path.isfile(output_file):
                    with open(output_file, "w", newline="") as f:
                        f.write("\t".join(header) + "\n")
                        f.write("\t".join([str(r) for r in row]) + "\n")
                else:
                    with open(output_file, "a", newline="") as f:
                        f.write("\t".join([str(r) for r in row]) + "\n")

    results = pd.DataFrame(
        {
            "lags": lags_list,
            "lags_label": lags_label_list,
            "params": params_list,
            **metric_dict,
        }
    )

    if results.empty:
        warnings.warn(
            "All models failed to train. Check the parameters and data.",
            RuntimeWarning,
        )
        return results

    results = results.sort_values(
        by=list(metric_dict.keys())[0], ascending=True if is_regression else False
    ).reset_index(drop=True)
    results = pd.concat([results, results["params"].apply(pd.Series)], axis=1)

    if return_best:
        best_lags = results.loc[0, "lags"]
        best_params = results.loc[0, "params"]
        best_metric = results.loc[0, list(metric_dict.keys())[0]]

        # NOTE: Here we use the actual forecaster passed by the user
        forecaster.set_lags(best_lags)
        forecaster.set_params(**best_params)

        forecaster.fit(y=y, exog=exog, store_in_sample_residuals=True)

        print(
            f"`Forecaster` refitted using the best-found lags and parameters, "
            f"and the whole data set: \n"
            f"  Lags: {best_lags} \n"
            f"  Parameters: {best_params}\n"
            f"  {'Backtesting' if isinstance(cv, TimeSeriesFold) else 'One-step-ahead'} "
            f"metric: {best_metric}"
        )

    set_skforecast_warnings(suppress_warnings, action="default")

    return results


def grid_search_forecaster(
    forecaster: object,
    y: pd.Series,
    cv: TimeSeriesFold | OneStepAheadFold,
    param_grid: dict,
    metric: str | Callable | list[str | Callable],
    exog: pd.Series | pd.DataFrame | None = None,
    lags_grid: (
        list[int | list[int] | np.ndarray[int] | range[int]]
        | dict[str, list[int | list[int] | np.ndarray[int] | range[int]]]
        | None
    ) = None,
    return_best: bool = True,
    n_jobs: int | str = "auto",
    verbose: bool = False,
    show_progress: bool = True,
    suppress_warnings: bool = False,
    output_file: str | None = None,
) -> pd.DataFrame:
    """
    Exhaustive grid search over parameter values for a Forecaster.
    """

    param_grid = list(ParameterGrid(param_grid))

    results = _evaluate_grid_hyperparameters(
        forecaster=forecaster,
        y=y,
        cv=cv,
        param_grid=param_grid,
        metric=metric,
        exog=exog,
        lags_grid=lags_grid,
        return_best=return_best,
        n_jobs=n_jobs,
        verbose=verbose,
        show_progress=show_progress,
        suppress_warnings=suppress_warnings,
        output_file=output_file,
    )

    return results



"""Metrics for evaluating forecasting models.

This module provides various metric functions for evaluating forecasting performance,
including custom metrics like MASE, RMSSE, and probabilistic metrics like CRPS.

Examples:
    Using standard metrics::

        import numpy as np
        from spotforecast2.forecaster.metrics import _get_metric

        y_true = np.array([1, 2, 3, 4, 5])
        y_pred = np.array([1.1, 1.9, 3.2, 3.8, 5.1])

        # Get a metric function
        mse = _get_metric('mean_squared_error')
        error = mse(y_true, y_pred)

    Using scaled metrics::

        from spotforecast2.forecaster.metrics import mean_absolute_scaled_error

        y_train = np.array([1, 2, 3, 4, 5, 6, 7, 8])
        y_true = np.array([9, 10, 11])
        y_pred = np.array([8.8, 10.2, 10.9])

        mase = mean_absolute_scaled_error(y_true, y_pred, y_train)
"""

from __future__ import annotations
from typing import Callable
import numpy as np
import pandas as pd
import inspect
from functools import wraps
from sklearn.metrics import (
    mean_squared_error,
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_log_error,
    median_absolute_error,
    mean_pinball_loss,
    accuracy_score,
    balanced_accuracy_score,
    f1_score,
    precision_score,
    recall_score,
)


def _get_metric(metric: str) -> Callable:
    """Get the corresponding scikit-learn function to calculate the metric.

    Args:
        metric: Metric used to quantify the goodness of fit of the model.

    Returns:
        scikit-learn function to calculate the desired metric.

    Examples:
        >>> from spotforecast2.forecaster.metrics import _get_metric
        >>> mse_func = _get_metric('mean_squared_error')
        >>> y_true = np.array([1, 2, 3])
        >>> y_pred = np.array([1.1, 1.9, 3.2])
        >>> error = mse_func(y_true, y_pred)
        >>> error > 0
        True
    """

    allowed_metrics = [
        # Regression metrics
        "mean_squared_error",
        "mean_absolute_error",
        "mean_absolute_percentage_error",
        "mean_squared_log_error",
        "mean_absolute_scaled_error",
        "root_mean_squared_scaled_error",
        "median_absolute_error",
        "symmetric_mean_absolute_percentage_error",
        # Classification metrics
        "accuracy_score",
        "balanced_accuracy_score",
        "f1_score",
        "precision_score",
        "recall_score",
    ]

    if metric not in allowed_metrics:
        raise ValueError(f"Allowed metrics are: {allowed_metrics}. Got {metric}.")

    metrics = {
        "mean_squared_error": mean_squared_error,
        "mean_absolute_error": mean_absolute_error,
        "mean_absolute_percentage_error": mean_absolute_percentage_error,
        "mean_squared_log_error": mean_squared_log_error,
        "mean_absolute_scaled_error": mean_absolute_scaled_error,
        "root_mean_squared_scaled_error": root_mean_squared_scaled_error,
        "median_absolute_error": median_absolute_error,
        "symmetric_mean_absolute_percentage_error": symmetric_mean_absolute_percentage_error,
        "accuracy_score": accuracy_score,
        "balanced_accuracy_score": balanced_accuracy_score,
        "f1_score": f1_score,
        "precision_score": precision_score,
        "recall_score": recall_score,
    }

    metric = add_y_train_argument(metrics[metric])

    return metric


def add_y_train_argument(func: Callable) -> Callable:
    """Add `y_train` argument to a function if it is not already present.

    Args:
        func: Function to which the argument is added.

    Returns:
        Function with `y_train` argument added.

    Examples:
        >>> def my_metric(y_true, y_pred):
        ...     return np.mean(np.abs(y_true - y_pred))
        >>> enhanced_metric = add_y_train_argument(my_metric)
        >>> # Now the function accepts y_train parameter
        >>> result = enhanced_metric(np.array([1,2,3]), np.array([1,2,3]), y_train=None)
    """

    sig = inspect.signature(func)

    if "y_train" in sig.parameters:
        return func

    new_params = list(sig.parameters.values()) + [
        inspect.Parameter("y_train", inspect.Parameter.KEYWORD_ONLY, default=None)
    ]
    new_sig = sig.replace(parameters=new_params)

    @wraps(func)
    def wrapper(*args, y_train=None, **kwargs):
        return func(*args, **kwargs)

    wrapper.__signature__ = new_sig

    return wrapper


def mean_absolute_scaled_error(
    y_true: np.ndarray | pd.Series,
    y_pred: np.ndarray | pd.Series,
    y_train: list[float] | np.ndarray | pd.Series,
) -> float:
    """Mean Absolute Scaled Error (MASE).

    MASE is a scale-independent error metric that measures the accuracy of
    a forecast. It is the mean absolute error of the forecast divided by the
    mean absolute error of a naive forecast in the training set. The naive
    forecast is the one obtained by shifting the time series by one period.
    If y_train is a list of numpy arrays or pandas Series, it is considered
    that each element is the true value of the target variable in the training
    set for each time series. In this case, the naive forecast is calculated
    for each time series separately.

    Args:
        y_true: True values of the target variable.
        y_pred: Predicted values of the target variable.
        y_train: True values of the target variable in the training set. If `list`, it
            is consider that each element is the true value of the target variable
            in the training set for each time series.

    Returns:
        MASE value.

    Examples:
        >>> from spotforecast2.forecaster.metrics import mean_absolute_scaled_error
        >>> y_train = np.array([1, 2, 3, 4, 5, 6, 7, 8])
        >>> y_true = np.array([9, 10, 11])
        >>> y_pred = np.array([8.8, 10.2, 10.9])
        >>> mase = mean_absolute_scaled_error(y_true, y_pred, y_train)
        >>> mase < 1.0  # Good forecast
        True
    """

    # NOTE: When using this metric in validation, `y_train` doesn't include
    # the first window_size observations used to create the predictors and/or
    # rolling features.

    if not isinstance(y_true, (pd.Series, np.ndarray)):
        raise TypeError("`y_true` must be a pandas Series or numpy ndarray.")
    if not isinstance(y_pred, (pd.Series, np.ndarray)):
        raise TypeError("`y_pred` must be a pandas Series or numpy ndarray.")
    if not isinstance(y_train, (list, pd.Series, np.ndarray)):
        raise TypeError("`y_train` must be a list, pandas Series or numpy ndarray.")
    if isinstance(y_train, list):
        for x in y_train:
            if not isinstance(x, (pd.Series, np.ndarray)):
                raise TypeError(
                    "When `y_train` is a list, each element must be a pandas Series "
                    "or numpy ndarray."
                )
    if len(y_true) != len(y_pred):
        raise ValueError("`y_true` and `y_pred` must have the same length.")
    if len(y_true) == 0 or len(y_pred) == 0:
        raise ValueError("`y_true` and `y_pred` must have at least one element.")

    if isinstance(y_train, list):
        # Flatten list of arrays for naive forecast if meaningful, but MASE usually assumes
        # naive forecast on single series. If list, we might be doing something else.
        # Original code does: np.concatenate([np.diff(x) for x in y_train])
        # This assumes independent series and we average error over all of them.
        naive_forecast = np.concatenate([np.diff(x) for x in y_train])
    else:
        naive_forecast = np.diff(y_train)

    mase = np.mean(np.abs(y_true - y_pred)) / np.nanmean(np.abs(naive_forecast))

    return mase


def root_mean_squared_scaled_error(
    y_true: np.ndarray | pd.Series,
    y_pred: np.ndarray | pd.Series,
    y_train: list[float] | np.ndarray | pd.Series,
) -> float:
    """Root Mean Squared Scaled Error (RMSSE).

    RMSSE is a scale-independent error metric that measures the accuracy of
    a forecast. It is the root mean squared error of the forecast divided by
    the root mean squared error of a naive forecast in the training set. The
    naive forecast is the one obtained by shifting the time series by one period.
    If y_train is a list of numpy arrays or pandas Series, it is considered
    that each element is the true value of the target variable in the training
    set for each time series. In this case, the naive forecast is calculated
    for each time series separately.

    Args:
        y_true: True values of the target variable.
        y_pred: Predicted values of the target variable.
        y_train: True values of the target variable in the training set. If list, it
            is consider that each element is the true value of the target variable
            in the training set for each time series.

    Returns:
        RMSSE value.

    Examples:
        >>> from spotforecast2.forecaster.metrics import root_mean_squared_scaled_error
        >>> y_train = np.array([1, 2, 3, 4, 5, 6, 7, 8])
        >>> y_true = np.array([9, 10, 11])
        >>> y_pred = np.array([8.8, 10.2, 10.9])
        >>> rmsse = root_mean_squared_scaled_error(y_true, y_pred, y_train)
        >>> rmsse < 1.0  # Good forecast
        True
    """

    # NOTE: When using this metric in validation, `y_train` doesn't include
    # the first window_size observations used to create the predictors and/or
    # rolling features.

    if not isinstance(y_true, (pd.Series, np.ndarray)):
        raise TypeError("`y_true` must be a pandas Series or numpy ndarray.")
    if not isinstance(y_pred, (pd.Series, np.ndarray)):
        raise TypeError("`y_pred` must be a pandas Series or numpy ndarray.")
    if not isinstance(y_train, (list, pd.Series, np.ndarray)):
        raise TypeError("`y_train` must be a list, pandas Series or numpy ndarray.")
    if isinstance(y_train, list):
        for x in y_train:
            if not isinstance(x, (pd.Series, np.ndarray)):
                raise TypeError(
                    "When `y_train` is a list, each element must be a pandas Series "
                    "or numpy ndarray."
                )
    if len(y_true) != len(y_pred):
        raise ValueError("`y_true` and `y_pred` must have the same length.")
    if len(y_true) == 0 or len(y_pred) == 0:
        raise ValueError("`y_true` and `y_pred` must have at least one element.")

    if isinstance(y_train, list):
        naive_forecast = np.concatenate([np.diff(x) for x in y_train])
    else:
        naive_forecast = np.diff(y_train)

    rmsse = np.sqrt(np.mean((y_true - y_pred) ** 2)) / np.sqrt(
        np.nanmean(naive_forecast**2)
    )

    return rmsse


def crps_from_predictions(y_true: float, y_pred: np.ndarray) -> float:
    """Compute the Continuous Ranked Probability Score (CRPS) from predictions.

    The CRPS compares the empirical distribution of a set of forecasted values
    to a scalar observation. The smaller the CRPS, the better.

    Args:
        y_true: The true value of the random variable.
        y_pred: The predicted values of the random variable. These are the multiple
            forecasted values for a single observation.

    Returns:
        The CRPS score.

    Examples:
        >>> from spotforecast2.forecaster.metrics import crps_from_predictions
        >>> y_true = 5.0
        >>> y_pred = np.array([4.5, 5.1, 4.9, 5.3, 4.7])
        >>> crps = crps_from_predictions(y_true, y_pred)
        >>> crps >= 0
        True
    """
    if not isinstance(y_pred, np.ndarray) or y_pred.ndim != 1:
        raise TypeError("`y_pred` must be a 1D numpy array.")

    if not isinstance(y_true, (float, int)):
        raise TypeError("`y_true` must be a float or integer.")

    y_pred = np.sort(y_pred)
    # Define the grid for integration including the true value
    grid = np.concatenate(([y_true], y_pred))
    grid = np.sort(grid)
    cdf_values = np.searchsorted(y_pred, grid, side="right") / len(y_pred)
    indicator = grid >= y_true
    diffs = np.diff(grid)
    crps = np.sum(diffs * (cdf_values[:-1] - indicator[:-1]) ** 2)

    return crps


def crps_from_quantiles(
    y_true: float,
    pred_quantiles: np.ndarray,
    quantile_levels: np.ndarray,
) -> float:
    """Calculate the Continuous Ranked Probability Score (CRPS) from quantiles.

    The empirical cdf is approximated using linear interpolation
    between the predicted quantiles.

    Args:
        y_true: The true value of the random variable.
        pred_quantiles: The predicted quantile values.
        quantile_levels: The quantile levels corresponding to the predicted quantiles.

    Returns:
        The CRPS score.

    Examples:
        >>> from spotforecast2.forecaster.metrics import crps_from_quantiles
        >>> y_true = 5.0
        >>> pred_quantiles = np.array([4.0, 4.5, 5.0, 5.5, 6.0])
        >>> quantile_levels = np.array([0.1, 0.25, 0.5, 0.75, 0.9])
        >>> crps = crps_from_quantiles(y_true, pred_quantiles, quantile_levels)
        >>> crps >= 0
        True
    """
    if not isinstance(y_true, (float, int)):
        raise TypeError("`y_true` must be a float or integer.")

    if not isinstance(pred_quantiles, np.ndarray) or pred_quantiles.ndim != 1:
        raise TypeError("`pred_quantiles` must be a 1D numpy array.")

    if not isinstance(quantile_levels, np.ndarray) or quantile_levels.ndim != 1:
        raise TypeError("`quantile_levels` must be a 1D numpy array.")

    if len(pred_quantiles) != len(quantile_levels):
        raise ValueError(
            "The number of predicted quantiles and quantile levels must be equal."
        )

    sorted_indices = np.argsort(pred_quantiles)
    pred_quantiles = pred_quantiles[sorted_indices]
    quantile_levels = quantile_levels[sorted_indices]

    # Define the empirical CDF function using interpolation
    def empirical_cdf(x):
        return np.interp(x, pred_quantiles, quantile_levels, left=0.0, right=1.0)

    # Define the CRPS integrand
    def crps_integrand(x):
        return (empirical_cdf(x) - (x >= y_true)) ** 2

    # Integration bounds: Extend slightly beyond predicted quantiles
    xmin = np.min(pred_quantiles) * 0.9
    xmax = np.max(pred_quantiles) * 1.1

    # Create a fine grid of x values for integration
    x_values = np.linspace(xmin, xmax, 1000)

    # Compute the integrand values and integrate using the trapezoidal rule
    integrand_values = crps_integrand(x_values)
    if np.__version__ >= "2.0.0":
        crps = np.trapezoid(integrand_values, x=x_values)
    else:
        crps = np.trapz(integrand_values, x_values)

    return crps


def calculate_coverage(
    y_true: np.ndarray | pd.Series,
    lower_bound: np.ndarray | pd.Series,
    upper_bound: np.ndarray | pd.Series,
) -> float:
    """Calculate coverage of a given interval.

    Coverage is the proportion of true values that fall within the interval.

    Args:
        y_true: True values of the target variable.
        lower_bound: Lower bound of the interval.
        upper_bound: Upper bound of the interval.

    Returns:
        Coverage of the interval.

    Examples:
        >>> from spotforecast2.forecaster.metrics import calculate_coverage
        >>> y_true = np.array([1, 2, 3, 4, 5])
        >>> lower_bound = np.array([0.5, 1.5, 2.5, 3.5, 4.5])
        >>> upper_bound = np.array([1.5, 2.5, 3.5, 4.5, 5.5])
        >>> coverage = calculate_coverage(y_true, lower_bound, upper_bound)
        >>> coverage == 1.0  # All values within bounds
        True
    """
    if not isinstance(y_true, (np.ndarray, pd.Series)) or y_true.ndim != 1:
        raise TypeError("`y_true` must be a 1D numpy array or pandas Series.")

    if not isinstance(lower_bound, (np.ndarray, pd.Series)) or lower_bound.ndim != 1:
        raise TypeError("`lower_bound` must be a 1D numpy array or pandas Series.")

    if not isinstance(upper_bound, (np.ndarray, pd.Series)) or upper_bound.ndim != 1:
        raise TypeError("`upper_bound` must be a 1D numpy array or pandas Series.")

    y_true = np.asarray(y_true)
    lower_bound = np.asarray(lower_bound)
    upper_bound = np.asarray(upper_bound)

    if y_true.shape != lower_bound.shape or y_true.shape != upper_bound.shape:
        raise ValueError(
            "`y_true`, `lower_bound` and `upper_bound` must have the same shape."
        )

    coverage = np.mean(np.logical_and(y_true >= lower_bound, y_true <= upper_bound))

    return coverage


def create_mean_pinball_loss(alpha: float) -> callable:
    """Create pinball loss for a given quantile.

    Also known as quantile loss. Internally, it uses the `mean_pinball_loss`
    function from scikit-learn.

    Args:
        alpha: Quantile for which the Pinball loss is calculated.
            Must be between 0 and 1, inclusive.

    Returns:
        Mean Pinball loss function for the given quantile.

    Examples:
        >>> from spotforecast2.forecaster.metrics import create_mean_pinball_loss
        >>> pinball_loss_50 = create_mean_pinball_loss(alpha=0.5)
        >>> y_true = np.array([1, 2, 3, 4, 5])
        >>> y_pred = np.array([1.1, 1.9, 3.2, 3.8, 5.1])
        >>> loss = pinball_loss_50(y_true, y_pred)
        >>> loss >= 0
        True
    """
    if not (0 <= alpha <= 1):
        raise ValueError("alpha must be between 0 and 1, both inclusive.")

    def mean_pinball_loss_q(y_true, y_pred):
        return mean_pinball_loss(y_true, y_pred, alpha=alpha)

    return mean_pinball_loss_q


def symmetric_mean_absolute_percentage_error(
    y_true: np.ndarray | pd.Series, y_pred: np.ndarray | pd.Series
) -> float:
    """Compute the Symmetric Mean Absolute Percentage Error (SMAPE).

    SMAPE is a relative error metric used to measure the accuracy
    of forecasts. Unlike MAPE, it is symmetric and prevents division
    by zero by averaging the absolute values of actual and predicted values.

    The result is expressed as a percentage and ranges from 0%
    (perfect prediction) to 200% (maximum error).

    Args:
        y_true: True values of the target variable.
        y_pred: Predicted values of the target variable.

    Returns:
        SMAPE value as a percentage.

    Examples:
        >>> from spotforecast2.forecaster.metrics import symmetric_mean_absolute_percentage_error
        >>> y_true = np.array([100, 200, 0])
        >>> y_pred = np.array([110, 180, 10])
        >>> result = symmetric_mean_absolute_percentage_error(y_true, y_pred)
        >>> 0 <= result <= 200
        True
    """

    if not isinstance(y_true, (pd.Series, np.ndarray)):
        raise TypeError("`y_true` must be a pandas Series or numpy ndarray.")
    if not isinstance(y_pred, (pd.Series, np.ndarray)):
        raise TypeError("`y_pred` must be a pandas Series or numpy ndarray.")
    if len(y_true) != len(y_pred):
        raise ValueError("`y_true` and `y_pred` must have the same length.")
    if len(y_true) == 0 or len(y_pred) == 0:
        raise ValueError("`y_true` and `y_pred` must have at least one element.")

    numerator = np.abs(y_true - y_pred)
    denominator = (np.abs(y_true) + np.abs(y_pred)) / 2

    # NOTE: Avoid division by zero
    mask = denominator != 0
    smape_values = np.zeros_like(denominator)
    smape_values[mask] = numerator[mask] / denominator[mask]

    smape = 100 * np.mean(smape_values)

    return smape



"""Custom exceptions and warnings for spotforecast2.

This module contains all the custom warnings and error classes used
across spotforecast2.

Examples:
    Using custom warnings::

        import warnings
        from spotforecast2.exceptions import MissingValuesWarning

        # Raise a warning
        warnings.warn(
            "Missing values detected in input data.",
            MissingValuesWarning
        )

        # Suppress a specific warning
        warnings.simplefilter('ignore', category=MissingValuesWarning)
"""

import warnings
import inspect
from functools import wraps
import textwrap

try:
    from rich.console import Console
    from rich.panel import Panel
    from rich.text import Text

    HAS_RICH = True
except ImportError:
    HAS_RICH = False


def runtime_deprecated(
    replacement: str = None,
    version: str = None,
    removal: str = None,
    category: type[Warning] = FutureWarning,
) -> object:
    """Decorator to mark functions or classes as deprecated.

    Works for both function and class targets, and ensures warnings are visible
    even inside Jupyter notebooks.

    Args:
        replacement: Name of the replacement function/class to use instead.
        version: Version in which the function/class was deprecated.
        removal: Version in which the function/class will be removed.
        category: Warning category to use. Default is FutureWarning.

    Returns:
        Decorator function.

    Examples:
        >>> @runtime_deprecated(replacement='new_function', version='0.5', removal='1.0')
        ... def old_function():
        ...     pass
        >>> old_function()  # doctest: +SKIP
        FutureWarning: old_function() is deprecated since version 0.5; use new_function instead...
    """

    def decorator(obj):
        is_function = inspect.isfunction(obj) or inspect.ismethod(obj)
        is_class = inspect.isclass(obj)

        if not (is_function or is_class):
            raise TypeError(
                "@runtime_deprecated can only be used on functions or classes"
            )

        # ----- Build warning message -----
        name = obj.__name__
        message = (
            f"{name}() is deprecated" if is_function else f"{name} class is deprecated"
        )
        if version:
            message += f" since version {version}"
        if replacement:
            message += f"; use {replacement} instead"
        if removal:
            message += f". It will be removed in version {removal}."
        else:
            message += "."

        def issue_warning():
            """Emit warning in a way that always shows in notebooks."""
            with warnings.catch_warnings():
                warnings.simplefilter("always", category)
                warnings.warn(message, category, stacklevel=3)

        # ----- Case 1: decorating a function -----
        if is_function:

            @wraps(obj)
            def wrapper(*args, **kwargs):
                issue_warning()
                return obj(*args, **kwargs)

            # Add metadata
            wrapper.__deprecated__ = True
            wrapper.__replacement__ = replacement
            wrapper.__version__ = version
            wrapper.__removal__ = removal
            return wrapper

        # ----- Case 2: decorating a class -----
        elif is_class:
            orig_init = getattr(obj, "__init__", None)
            orig_new = getattr(obj, "__new__", None)

            # Only wrap whichever exists (some classes use __new__, others __init__)
            if orig_new and (orig_new is not object.__new__):

                @wraps(orig_new)
                def wrapped_new(cls, *args, **kwargs):
                    issue_warning()
                    return orig_new(cls, *args, **kwargs)

                obj.__new__ = staticmethod(wrapped_new)

            elif orig_init:

                @wraps(orig_init)
                def wrapped_init(self, *args, **kwargs):
                    issue_warning()
                    return orig_init(self, *args, **kwargs)

                obj.__init__ = wrapped_init

            # Add metadata
            obj.__deprecated__ = True
            obj.__replacement__ = replacement
            obj.__version__ = version
            obj.__removal__ = removal

            return obj

    return decorator


class DataTypeWarning(UserWarning):
    """Warning for incompatible data types in exogenous data.

    Used to notify there are dtypes in the exogenous data that are not
    'int', 'float', 'bool' or 'category'. Most machine learning models do not
    accept other data types, therefore the forecaster `fit` and `predict` may fail.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Exogenous data contains unsupported dtypes.",
        ...     DataTypeWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=DataTypeWarning)"
        )
        return self.message + "\\n" + extra_message


class DataTransformationWarning(UserWarning):
    """Warning for output data in transformed space.

    Used to notify that the output data is in the transformed space.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Output is in transformed space.",
        ...     DataTransformationWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=DataTransformationWarning)"
        )
        return self.message + "\\n" + extra_message


class ExogenousInterpretationWarning(UserWarning):
    """Warning about implications when using exogenous variables.

    Used to notify about important implications when using exogenous
    variables with models that use a two-step approach (e.g., regression + ARAR).

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Exogenous variables may not be used as expected.",
        ...     ExogenousInterpretationWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=ExogenousInterpretationWarning)"
        )
        return self.message + "\\n" + extra_message


class FeatureOutOfRangeWarning(UserWarning):
    """Warning for features out of training range.

    Used to notify that a feature is out of the range seen during training.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Feature value exceeds training range.",
        ...     FeatureOutOfRangeWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=FeatureOutOfRangeWarning)"
        )
        return self.message + "\\n" + extra_message


class IgnoredArgumentWarning(UserWarning):
    """Warning for ignored arguments.

    Used to notify that an argument is ignored when using a method
    or a function.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Argument 'x' is ignored in this context.",
        ...     IgnoredArgumentWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=IgnoredArgumentWarning)"
        )
        return self.message + "\\n" + extra_message


class InputTypeWarning(UserWarning):
    """Warning for inefficient input format.

    Used to notify that input format is not the most efficient or
    recommended for the forecaster.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Input format is not optimal for this forecaster.",
        ...     InputTypeWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=InputTypeWarning)"
        )
        return self.message + "\\n" + extra_message


class LongTrainingWarning(UserWarning):
    """Warning for potentially long training processes.

    Used to notify that a large number of models will be trained and the
    the process may take a while to run.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Training may take a long time.",
        ...     LongTrainingWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=LongTrainingWarning)"
        )
        return self.message + "\\n" + extra_message


class MissingExogWarning(UserWarning):
    """Warning for missing exogenous variables.

    Used to indicate that there are missing exogenous variables in the
    data. Most machine learning models do not accept missing values, so the
    Forecaster's `fit' and `predict' methods may fail.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Missing exogenous variables detected.",
        ...     MissingExogWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=MissingExogWarning)"
        )
        return self.message + "\\n" + extra_message


class MissingValuesWarning(UserWarning):
    """Warning for missing values in data.

    Used to indicate that there are missing values in the data. This
    warning occurs when the input data contains missing values, or the training
    matrix generates missing values. Most machine learning models do not accept
    missing values, so the Forecaster's `fit' and `predict' methods may fail.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Missing values detected in input data.",
        ...     MissingValuesWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=MissingValuesWarning)"
        )
        return self.message + "\\n" + extra_message


class OneStepAheadValidationWarning(UserWarning):
    """Warning for one-step-ahead validation usage.

    Used to notify that the one-step-ahead validation is being used.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Using one-step-ahead validation.",
        ...     OneStepAheadValidationWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=OneStepAheadValidationWarning)"
        )
        return self.message + "\\n" + extra_message


class ResidualsUsageWarning(UserWarning):
    """Warning for incorrect residuals usage.

    Used to notify that a residual are not correctly used in the
    probabilistic forecasting process.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Residuals are not properly used.",
        ...     ResidualsUsageWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=ResidualsUsageWarning)"
        )
        return self.message + "\\n" + extra_message


class UnknownLevelWarning(UserWarning):
    """Warning for unknown levels in prediction.

    Used to notify that a level being predicted was not part of the
    training data.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Predicting for an unknown level.",
        ...     UnknownLevelWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=UnknownLevelWarning)"
        )
        return self.message + "\\n" + extra_message


class SaveLoadSkforecastWarning(UserWarning):
    """Warning for save/load operations.

    Used to notify any issues that may arise when saving or loading
    a forecaster.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Issues detected when saving forecaster.",
        ...     SaveLoadSkforecastWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=SaveLoadSkforecastWarning)"
        )
        return self.message + "\\n" + extra_message


class SpotforecastVersionWarning(UserWarning):
    """Warning for version mismatch.

    Used to notify that the version installed in the
    environment differs from the version used to initialize the forecaster.

    Examples:
        >>> import warnings
        >>> warnings.warn(
        ...     "Version mismatch detected.",
        ...     SpotforecastVersionWarning
        ... )  # doctest: +SKIP
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        extra_message = (
            "You can suppress this warning using: "
            "warnings.simplefilter('ignore', category=SpotforecastVersionWarning)"
        )
        return self.message + "\\n" + extra_message


class NotFittedError(ValueError, AttributeError):
    """Exception class to raise if estimator is used before fitting.

    This class inherits from both ValueError and AttributeError to help with
    exception handling and backward compatibility.

    Examples:
        >>> from spotforecast2.exceptions import NotFittedError
        >>> try:
        ...     raise NotFittedError("Forecaster not fitted")
        ... except NotFittedError as e:
        ...     print(e)
        Forecaster not fitted
    """


warn_skforecast_categories = [
    DataTypeWarning,
    DataTransformationWarning,
    ExogenousInterpretationWarning,
    FeatureOutOfRangeWarning,
    IgnoredArgumentWarning,
    InputTypeWarning,
    LongTrainingWarning,
    MissingExogWarning,
    MissingValuesWarning,
    OneStepAheadValidationWarning,
    ResidualsUsageWarning,
    UnknownLevelWarning,
    SaveLoadSkforecastWarning,
    SpotforecastVersionWarning,
]


def format_warning_handler(
    message: str,
    category: str,
    filename: str,
    lineno: str,
    file: object = None,
    line: str = None,
) -> None:
    """Custom warning handler to format warnings in a box.

    Args:
        message: Warning message.
        category: Warning category.
        filename: Filename where the warning was raised.
        lineno: Line number where the warning was raised.
        file: File where the warning was raised.
        line: Line where the warning was raised.

    Returns:
        None

    Examples:
        >>> # This is used internally by the warnings module
        >>> set_warnings_style('skforecast')  # doctest: +SKIP
    """

    if isinstance(message, tuple(warn_skforecast_categories)):
        width = 88
        title = type(message).__name__
        output_text = ["\\n"]

        wrapped_message = textwrap.fill(
            str(message), width=width - 2, expand_tabs=True, replace_whitespace=True
        )
        title_top_border = f"╭{'─' * ((width - len(title) - 2) // 2)} {title} {'─' * ((width - len(title) - 2) // 2)}╮"
        if len(title) % 2 != 0:
            title_top_border = title_top_border[:-1] + "─" + "╮"
        bottom_border = f"╰{'─' * width}╯"
        output_text.append(title_top_border)

        for line in wrapped_message.split("\\n"):
            output_text.append(f"│ {line.ljust(width - 2)} │")

        output_text.append(bottom_border)
        output_text = "\\n".join(output_text)
        color = "\\033[38;5;208m"
        reset = "\\033[0m"
        output_text = f"{color}{output_text}{reset}"
        print(output_text)
    else:
        # Fallback to default Python warning formatting
        warnings._original_showwarning(message, category, filename, lineno, file, line)


def rich_warning_handler(
    message: str,
    category: str,
    filename: str,
    lineno: str,
    file: object = None,
    line: str = None,
) -> None:
    """Custom handler for warnings that uses rich to display formatted panels.

    Args:
        message: Warning message.
        category: Warning category.
        filename: Filename where the warning was raised.
        lineno: Line number where the warning was raised.
        file: File where the warning was raised.
        line: Line where the warning was raised.

    Returns:
        None

    Examples:
        >>> # This is used internally when rich is available
        >>> set_warnings_style('skforecast')  # doctest: +SKIP
    """

    if isinstance(message, tuple(warn_skforecast_categories)):
        if not HAS_RICH:
            # Fallback to format_warning_handler if rich is not available
            format_warning_handler(message, category, filename, lineno, file, line)
            return

        console = Console()

        category_name = category.__name__
        text = (
            f"{message.message}\\n\\n"
            f"Category : spotforecast2.exceptions.{category_name}\\n"
            f"Location : {filename}:{lineno}\\n"
            f"Suppress : warnings.simplefilter('ignore', category={category_name})"
        )

        panel = Panel(
            Text(text, justify="left"),
            title=category_name,
            title_align="center",
            border_style="color(214)",
            width=88,
        )

        console.print(panel)
    else:
        # Fallback to default Python warning formatting
        warnings._original_showwarning(message, category, filename, lineno, file, line)


def set_warnings_style(style: str = "skforecast") -> None:
    """Set the warning handler based on the provided style.

    Args:
        style: The style of the warning handler. Either 'skforecast' or 'default'.

    Returns:
        None

    Examples:
        >>> set_warnings_style('skforecast')
        >>> # Now warnings will be displayed with formatting
        >>> set_warnings_style('default')
        >>> # Back to default Python warning format
    """
    if style == "skforecast":
        if not hasattr(warnings, "_original_showwarning"):
            warnings._original_showwarning = warnings.showwarning
        if HAS_RICH:
            warnings.showwarning = rich_warning_handler
        else:
            warnings.showwarning = format_warning_handler
    else:
        if hasattr(warnings, "_original_showwarning"):
            warnings.showwarning = warnings._original_showwarning


set_warnings_style(style="skforecast")


def set_skforecast_warnings(suppress_warnings: bool, action: str = "ignore") -> None:
    """
    Suppress spotforecast warnings.

    Args:
        suppress_warnings: bool
            If True, spotforecast warnings will be suppressed.
        action: str, default 'ignore'
            Action to take regarding the warnings.
    """
    if suppress_warnings:
        for category in warn_skforecast_categories:
            warnings.simplefilter(action, category=category)



"""Metrics calculation utilities for model selection."""

from __future__ import annotations
import pandas as pd


def _calculate_metrics_one_step_ahead(
    forecaster: object,
    metrics: list,
    X_train: pd.DataFrame,
    y_train: pd.Series | dict[int, pd.Series],
    X_test: pd.DataFrame,
    y_test: pd.Series | dict[int, pd.Series],
) -> list:
    """
    Calculate metrics when predictions are one-step-ahead.

    When forecaster is of type `ForecasterDirect`, only the estimator for step 1
    is used.

    Args:
        forecaster: Forecaster model.
        metrics: List of metrics.
        X_train: Predictor values used to train the model.
        y_train: Target values related to each row of `X_train`.
        X_test: Predictor values used to test the model.
        y_test: Target values related to each row of `X_test`.

    Returns:
        List with metric values.

    Notes:
        When using this metric in validation, `y_train` doesn't include
        the first window_size observations used to create the predictors and/or
        rolling features.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> from sklearn.linear_model import LinearRegression
        >>> from sklearn.metrics import mean_squared_error
        >>> from spotforecast2.forecaster.recursive import ForecasterRecursive
        >>> from spotforecast2.model_selection.utils_metrics import _calculate_metrics_one_step_ahead
        >>>
        >>> forecaster = ForecasterRecursive(LinearRegression(), lags=2)
        >>>
        >>> # Mock training data (already transformed into predictors/target)
        >>> X_train = pd.DataFrame(np.array([[1, 2], [2, 3], [3, 4]]), columns=['lag_2', 'lag_1'])
        >>> y_train = pd.Series(np.array([3, 4, 5]), name='y')
        >>>
        >>> # Mock test data
        >>> X_test = pd.DataFrame(np.array([[4, 5]]), columns=['lag_2', 'lag_1'])
        >>> y_test = pd.Series(np.array([6]), name='y')
        >>>
        >>> metrics = [mean_squared_error]
        >>> result = _calculate_metrics_one_step_ahead(
        ...     forecaster=forecaster,
        ...     metrics=metrics,
        ...     X_train=X_train,
        ...     y_train=y_train,
        ...     X_test=X_test,
        ...     y_test=y_test
        ... )
        >>> print(result)
        [0.0]
    """

    if type(forecaster).__name__ == "ForecasterDirect":

        step = 1  # Only the model for step 1 is optimized.
        X_train, y_train = forecaster.filter_train_X_y_for_step(
            step=step, X_train=X_train, y_train=y_train
        )
        X_test, y_test = forecaster.filter_train_X_y_for_step(
            step=step, X_train=X_test, y_train=y_test
        )
        forecaster.estimators_[step].fit(X_train, y_train)
        y_pred = forecaster.estimators_[step].predict(X_test)

    else:
        forecaster.estimator.fit(X_train, y_train)
        y_pred = forecaster.estimator.predict(X_test)

    y_true = y_test.to_numpy()
    y_pred = y_pred.ravel()
    y_train = y_train.to_numpy()

    if forecaster.differentiation is not None:
        y_true = forecaster.differentiator.inverse_transform_next_window(y_true)
        y_pred = forecaster.differentiator.inverse_transform_next_window(y_pred)
        y_train = forecaster.differentiator.inverse_transform_training(y_train)

    if forecaster.transformer_y is not None:
        y_true = forecaster.transformer_y.inverse_transform(y_true.reshape(-1, 1))
        y_pred = forecaster.transformer_y.inverse_transform(y_pred.reshape(-1, 1))
        y_train = forecaster.transformer_y.inverse_transform(y_train.reshape(-1, 1))

    y_true = y_true.ravel()
    y_pred = y_pred.ravel()
    y_train = y_train.ravel()
    metric_values = [m(y_true=y_true, y_pred=y_pred, y_train=y_train) for m in metrics]

    return metric_values
